from __future__ import annotations

import json
import os
from pathlib import Path
from typing import TYPE_CHECKING, Any

import pytest
import yaml

from commitizen import cmd, commands
from commitizen.__version__ import __version__
from commitizen.exceptions import InitFailedError, NoAnswersError

if TYPE_CHECKING:
    from pytest_mock import MockFixture

    from commitizen.config.base_config import BaseConfig


class FakeQuestion:
    def __init__(self, expected_return):
        self.expected_return = expected_return

    def ask(self):
        return self.expected_return

    def unsafe_ask(self):
        return self.expected_return


pre_commit_config_filename = ".pre-commit-config.yaml"
cz_hook_config = {
    "repo": "https://github.com/commitizen-tools/commitizen",
    "rev": f"v{__version__}",
    "hooks": [
        {"id": "commitizen"},
        {"id": "commitizen-branch", "stages": ["pre-push"]},
    ],
}

expected_config = (
    "[tool.commitizen]\n"
    'name = "cz_conventional_commits"\n'
    'tag_format = "$version"\n'
    'version_scheme = "semver"\n'
    'version = "0.0.1"\n'
    "update_changelog_on_bump = true\n"
    "major_version_zero = true\n"
)

EXPECTED_DICT_CONFIG = {
    "commitizen": {
        "name": "cz_conventional_commits",
        "tag_format": "$version",
        "version_scheme": "semver",
        "version": "0.0.1",
        "update_changelog_on_bump": True,
        "major_version_zero": True,
    }
}


def test_init_without_setup_pre_commit_hook(
    tmpdir, mocker: MockFixture, config: BaseConfig
):
    mocker.patch(
        "questionary.select",
        side_effect=[
            FakeQuestion("pyproject.toml"),
            FakeQuestion("cz_conventional_commits"),
            FakeQuestion("commitizen"),
            FakeQuestion("semver"),
        ],
    )
    mocker.patch("questionary.confirm", return_value=FakeQuestion(True))
    mocker.patch("questionary.text", return_value=FakeQuestion("$version"))
    # Return None to skip hook installation
    mocker.patch("questionary.checkbox", return_value=FakeQuestion(None))
    with tmpdir.as_cwd():
        commands.Init(config)()

        with open("pyproject.toml", encoding="utf-8") as toml_file:
            config_data = toml_file.read()
        assert config_data == expected_config

        assert not os.path.isfile(pre_commit_config_filename)


def test_init_when_config_already_exists(config: BaseConfig, capsys):
    # Set config path
    path = Path(os.sep.join(["tests", "pyproject.toml"]))
    config.path = path

    commands.Init(config)()
    captured = capsys.readouterr()
    assert captured.out == f"Config file {path} already exists\n"


def test_init_without_choosing_tag(config: BaseConfig, mocker: MockFixture, tmpdir):
    mocker.patch(
        "commitizen.commands.init.get_tag_names", return_value=["0.0.2", "0.0.1"]
    )
    mocker.patch("commitizen.commands.init.get_latest_tag_name", return_value="0.0.2")
    mocker.patch(
        "questionary.select",
        side_effect=[
            FakeQuestion("pyproject.toml"),
            FakeQuestion("cz_conventional_commits"),
            FakeQuestion("commitizen"),
            FakeQuestion(""),
        ],
    )
    mocker.patch("questionary.confirm", return_value=FakeQuestion(False))
    mocker.patch("questionary.text", return_value=FakeQuestion("y"))

    with tmpdir.as_cwd():
        with pytest.raises(NoAnswersError):
            commands.Init(config)()


@pytest.fixture(scope="function")
def pre_commit_installed(mocker: MockFixture):
    # Assume the `pre-commit` is installed
    mocker.patch(
        "commitizen.project_info.is_pre_commit_installed",
        return_value=True,
    )
    # And installation success (i.e. no exception raised)
    mocker.patch(
        "commitizen.cmd.run",
        return_value=cmd.Command("0.0.1", "", b"", b"", 0),
    )


@pytest.fixture(scope="function", params=["pyproject.toml", ".cz.json", ".cz.yaml"])
def default_choice(request, mocker: MockFixture):
    mocker.patch(
        "questionary.select",
        side_effect=[
            FakeQuestion(request.param),
            FakeQuestion("cz_conventional_commits"),
            FakeQuestion("commitizen"),
            FakeQuestion("semver"),
        ],
    )
    mocker.patch("questionary.confirm", return_value=FakeQuestion(True))
    mocker.patch("questionary.text", return_value=FakeQuestion("$version"))
    mocker.patch(
        "questionary.checkbox",
        return_value=FakeQuestion(["commit-msg", "pre-push"]),
    )
    yield request.param


def check_cz_config(config_filepath: str):
    """
    Check the content of commitizen config is as expected
    """
    with open(config_filepath) as file:
        if "json" in config_filepath:
            assert json.load(file) == EXPECTED_DICT_CONFIG
        elif "yaml" in config_filepath:
            assert yaml.load(file, Loader=yaml.FullLoader) == EXPECTED_DICT_CONFIG
        else:
            config_data = file.read()
            assert config_data == expected_config


def check_pre_commit_config(expected: list[dict[str, Any]]):
    """
    Check the content of pre-commit config is as expected
    """
    with open(pre_commit_config_filename) as pre_commit_file:
        pre_commit_config_data = yaml.safe_load(pre_commit_file.read())
    assert pre_commit_config_data == {"repos": expected}


@pytest.mark.usefixtures("pre_commit_installed")
class TestPreCommitCases:
    def test_no_existing_pre_commit_config(
        _, default_choice: str, tmpdir, config: BaseConfig
    ):
        with tmpdir.as_cwd():
            commands.Init(config)()
            check_cz_config(default_choice)
            check_pre_commit_config([cz_hook_config])

    def test_empty_pre_commit_config(
        _, default_choice: str, tmpdir, config: BaseConfig
    ):
        with tmpdir.as_cwd():
            p = tmpdir.join(pre_commit_config_filename)
            p.write("")

            commands.Init(config)()
            check_cz_config(default_choice)
            check_pre_commit_config([cz_hook_config])

    def test_pre_commit_config_without_cz_hook(
        _, default_choice: str, tmpdir, config: BaseConfig
    ):
        existing_hook_config = {
            "repo": "https://github.com/pre-commit/pre-commit-hooks",
            "rev": "v1.2.3",
            "hooks": [{"id", "trailing-whitespace"}],
        }

        with tmpdir.as_cwd():
            p = tmpdir.join(pre_commit_config_filename)
            p.write(yaml.safe_dump({"repos": [existing_hook_config]}))

            commands.Init(config)()
            check_cz_config(default_choice)
            check_pre_commit_config([existing_hook_config, cz_hook_config])

    def test_cz_hook_exists_in_pre_commit_config(
        _, default_choice: str, tmpdir, config: BaseConfig
    ):
        with tmpdir.as_cwd():
            p = tmpdir.join(pre_commit_config_filename)
            p.write(yaml.safe_dump({"repos": [cz_hook_config]}))

            commands.Init(config)()
            check_cz_config(default_choice)
            # check that config is not duplicated
            check_pre_commit_config([cz_hook_config])


class TestNoPreCommitInstalled:
    def test_pre_commit_not_installed(
        _, mocker: MockFixture, config: BaseConfig, default_choice: str, tmpdir
    ):
        # Assume `pre-commit` is not installed
        mocker.patch(
            "commitizen.project_info.is_pre_commit_installed",
            return_value=False,
        )
        with tmpdir.as_cwd():
            with pytest.raises(InitFailedError):
                commands.Init(config)()


class TestAskTagFormat:
    def test_confirm_v_tag_format(self, mocker: MockFixture, config: BaseConfig):
        init = commands.Init(config)
        mocker.patch("questionary.confirm", return_value=FakeQuestion(True))

        result = init._ask_tag_format("v1.0.0")
        assert result == r"v$version"

    def test_reject_v_tag_format(self, mocker: MockFixture, config: BaseConfig):
        init = commands.Init(config)
        mocker.patch("questionary.confirm", return_value=FakeQuestion(False))
        mocker.patch("questionary.text", return_value=FakeQuestion("custom-$version"))

        result = init._ask_tag_format("v1.0.0")
        assert result == "custom-$version"

    def test_non_v_tag_format(self, mocker: MockFixture, config: BaseConfig):
        init = commands.Init(config)
        mocker.patch("questionary.text", return_value=FakeQuestion("custom-$version"))

        result = init._ask_tag_format("1.0.0")
        assert result == "custom-$version"

    def test_empty_input_returns_default(self, mocker: MockFixture, config: BaseConfig):
        init = commands.Init(config)
        mocker.patch("questionary.confirm", return_value=FakeQuestion(False))
        mocker.patch("questionary.text", return_value=FakeQuestion(""))

        result = init._ask_tag_format("v1.0.0")
        assert result == "$version"  # This is the default format from DEFAULT_SETTINGS


def test_init_with_confirmed_tag_format(
    config: BaseConfig, mocker: MockFixture, tmpdir
):
    mocker.patch(
        "commitizen.commands.init.get_tag_names", return_value=["v0.0.2", "v0.0.1"]
    )
    mocker.patch("commitizen.commands.init.get_latest_tag_name", return_value="v0.0.2")
    mocker.patch(
        "questionary.select",
        side_effect=[
            FakeQuestion("pyproject.toml"),
            FakeQuestion("cz_conventional_commits"),
            FakeQuestion("commitizen"),
            FakeQuestion("semver"),
        ],
    )
    mocker.patch("questionary.confirm", return_value=FakeQuestion(True))
    mocker.patch("questionary.text", return_value=FakeQuestion("$version"))
    mocker.patch("questionary.checkbox", return_value=FakeQuestion(None))

    with tmpdir.as_cwd():
        commands.Init(config)()
        with open("pyproject.toml", encoding="utf-8") as toml_file:
            assert 'tag_format = "v$version"' in toml_file.read()


def test_init_with_no_existing_tags(config: BaseConfig, mocker: MockFixture, tmpdir):
    mocker.patch("commitizen.commands.init.get_tag_names", return_value=[])
    mocker.patch("commitizen.commands.init.get_latest_tag_name", return_value="v1.0.0")
    mocker.patch(
        "questionary.select",
        side_effect=[
            FakeQuestion("pyproject.toml"),
            FakeQuestion("cz_conventional_commits"),
            FakeQuestion("commitizen"),
            FakeQuestion("semver"),
        ],
    )
    mocker.patch("questionary.confirm", return_value=FakeQuestion(False))
    mocker.patch("questionary.text", return_value=FakeQuestion("$version"))
    mocker.patch("questionary.checkbox", return_value=FakeQuestion(None))

    with tmpdir.as_cwd():
        commands.Init(config)()
        with open("pyproject.toml", encoding="utf-8") as toml_file:
            assert 'version = "0.0.1"' in toml_file.read()


def test_init_with_no_existing_latest_tag(
    config: BaseConfig, mocker: MockFixture, tmpdir
):
    mocker.patch("commitizen.commands.init.get_latest_tag_name", return_value=None)
    mocker.patch(
        "questionary.select",
        side_effect=[
            FakeQuestion("pyproject.toml"),
            FakeQuestion("cz_conventional_commits"),
            FakeQuestion("commitizen"),
            FakeQuestion("semver"),
        ],
    )
    mocker.patch("questionary.confirm", return_value=FakeQuestion(True))
    mocker.patch("questionary.text", return_value=FakeQuestion("$version"))
    mocker.patch("questionary.checkbox", return_value=FakeQuestion(None))

    with tmpdir.as_cwd():
        commands.Init(config)()
        with open("pyproject.toml", encoding="utf-8") as toml_file:
            assert 'version = "0.0.1"' in toml_file.read()


def test_init_with_existing_tags(config: BaseConfig, mocker: MockFixture, tmpdir):
    expected_tags = ["v1.0.0", "v0.9.0", "v0.8.0"]
    mocker.patch("commitizen.commands.init.get_tag_names", return_value=expected_tags)
    mocker.patch("commitizen.commands.init.get_latest_tag_name", return_value="v1.0.0")
    mocker.patch(
        "questionary.select",
        side_effect=[
            FakeQuestion("pyproject.toml"),
            FakeQuestion("cz_conventional_commits"),
            FakeQuestion("commitizen"),
            FakeQuestion("semver"),  # Select version scheme first
            FakeQuestion("v1.0.0"),  # Then select the latest tag
        ],
    )
    mocker.patch("questionary.confirm", return_value=FakeQuestion(True))
    mocker.patch("questionary.text", return_value=FakeQuestion("$version"))
    mocker.patch("questionary.checkbox", return_value=FakeQuestion(None))

    with tmpdir.as_cwd():
        commands.Init(config)()
        with open("pyproject.toml", encoding="utf-8") as toml_file:
            assert 'version = "1.0.0"' in toml_file.read()


def test_init_with_valid_tag_selection(config: BaseConfig, mocker: MockFixture, tmpdir):
    expected_tags = ["v1.0.0", "v0.9.0", "v0.8.0"]
    mocker.patch("commitizen.commands.init.get_tag_names", return_value=expected_tags)
    mocker.patch("commitizen.commands.init.get_latest_tag_name", return_value="v1.0.0")

    # Mock all questionary.select calls in the exact order they appear in Init.__call__
    mocker.patch(
        "questionary.select",
        side_effect=[
            FakeQuestion("pyproject.toml"),  # _ask_config_path
            FakeQuestion("cz_conventional_commits"),  # _ask_name
            FakeQuestion("commitizen"),  # _ask_version_provider
            FakeQuestion("v0.9.0"),  # _ask_tag (after confirm=False)
            FakeQuestion("semver"),  # _ask_version_scheme
        ],
    )

    mocker.patch(
        "questionary.confirm", return_value=FakeQuestion(False)
    )  # Don't confirm latest tag
    mocker.patch("questionary.text", return_value=FakeQuestion("$version"))
    mocker.patch("questionary.checkbox", return_value=FakeQuestion(None))

    with tmpdir.as_cwd():
        commands.Init(config)()
        with open("pyproject.toml", encoding="utf-8") as toml_file:
            content = toml_file.read()
            assert 'version = "0.9.0"' in content
            assert 'version_scheme = "semver"' in content


def test_init_configuration_settings(tmpdir, mocker: MockFixture, config: BaseConfig):
    """Test that all configuration settings are properly initialized."""
    mocker.patch(
        "questionary.select",
        side_effect=[
            FakeQuestion("pyproject.toml"),
            FakeQuestion("cz_conventional_commits"),
            FakeQuestion("commitizen"),
            FakeQuestion("semver"),
        ],
    )
    mocker.patch("questionary.confirm", return_value=FakeQuestion(True))
    mocker.patch("questionary.text", return_value=FakeQuestion("$version"))
    mocker.patch("questionary.checkbox", return_value=FakeQuestion(None))

    with tmpdir.as_cwd():
        commands.Init(config)()

        with open("pyproject.toml", encoding="utf-8") as toml_file:
            config_data = toml_file.read()

        # Verify all expected settings are present
        assert 'name = "cz_conventional_commits"' in config_data
        assert 'tag_format = "$version"' in config_data
        assert 'version_scheme = "semver"' in config_data
        assert 'version = "0.0.1"' in config_data
        assert "update_changelog_on_bump = true" in config_data
        assert "major_version_zero = true" in config_data


def test_init_configuration_with_version_provider(
    tmpdir, mocker: MockFixture, config: BaseConfig
):
    """Test configuration initialization with a different version provider."""
    mocker.patch(
        "questionary.select",
        side_effect=[
            FakeQuestion("pyproject.toml"),
            FakeQuestion("cz_conventional_commits"),
            FakeQuestion("pep621"),  # Different version provider
            FakeQuestion("semver"),
        ],
    )
    mocker.patch("questionary.confirm", return_value=FakeQuestion(True))
    mocker.patch("questionary.text", return_value=FakeQuestion("$version"))
    mocker.patch("questionary.checkbox", return_value=FakeQuestion(None))

    with tmpdir.as_cwd():
        commands.Init(config)()

        with open("pyproject.toml", encoding="utf-8") as toml_file:
            config_data = toml_file.read()

        # Verify version provider is set instead of version
        assert 'name = "cz_conventional_commits"' in config_data
        assert 'tag_format = "$version"' in config_data
        assert 'version_scheme = "semver"' in config_data
        assert 'version_provider = "pep621"' in config_data
        assert "update_changelog_on_bump = true" in config_data
        assert "major_version_zero = true" in config_data
        assert (
            "version = " not in config_data
        )  # Version should not be set when using version_provider



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, NamedTuple

import questionary
import yaml

from commitizen import cmd, factory, out, project_info
from commitizen.__version__ import __version__
from commitizen.config.factory import create_config
from commitizen.cz import registry
from commitizen.defaults import CONFIG_FILES, DEFAULT_SETTINGS
from commitizen.exceptions import InitFailedError, NoAnswersError
from commitizen.git import get_latest_tag_name, get_tag_names, smart_open
from commitizen.version_schemes import KNOWN_SCHEMES, Version, get_version_scheme

if TYPE_CHECKING:
    from commitizen.config import (
        BaseConfig,
    )


class _VersionProviderOption(NamedTuple):
    provider_name: str
    description: str

    @property
    def title(self) -> str:
        return f"{self.provider_name}: {self.description}"


_VERSION_PROVIDER_CHOICES = tuple(
    questionary.Choice(title=option.title, value=option.provider_name)
    for option in (
        _VersionProviderOption(
            provider_name="commitizen",
            description="Fetch and set version in commitizen config (default)",
        ),
        _VersionProviderOption(
            provider_name="cargo",
            description="Get and set version from Cargo.toml:project.version field",
        ),
        _VersionProviderOption(
            provider_name="composer",
            description="Get and set version from composer.json:project.version field",
        ),
        _VersionProviderOption(
            provider_name="npm",
            description="Get and set version from package.json:project.version field",
        ),
        _VersionProviderOption(
            provider_name="pep621",
            description="Get and set version from pyproject.toml:project.version field",
        ),
        _VersionProviderOption(
            provider_name="poetry",
            description="Get and set version from pyproject.toml:tool.poetry.version field",
        ),
        _VersionProviderOption(
            provider_name="uv",
            description="Get and set version from pyproject.toml and uv.lock",
        ),
        _VersionProviderOption(
            provider_name="scm",
            description="Fetch the version from git and does not need to set it back",
        ),
    )
)


class Init:
    _PRE_COMMIT_CONFIG_PATH = ".pre-commit-config.yaml"

    def __init__(self, config: BaseConfig, *args: object) -> None:
        self.config: BaseConfig = config
        self.cz = factory.committer_factory(self.config)

    def __call__(self) -> None:
        if self.config.path:
            out.line(f"Config file {self.config.path} already exists")
            return

        out.info("Welcome to commitizen!\n")
        out.line(
            "Answer the following questions to configure your project.\n"
            "For further configuration, visit:\n"
            "\n"
            "https://commitizen-tools.github.io/commitizen/config/"
            "\n"
        )

        # Collect information
        try:
            config_path = self._ask_config_path()  # select
            cz_name = self._ask_name()  # select
            version_provider = self._ask_version_provider()  # select
            tag = self._ask_tag()  # confirm & select
            version_scheme = self._ask_version_scheme()  # select
            version = get_version_scheme(self.config.settings, version_scheme)(tag)
            tag_format = self._ask_tag_format(tag)  # confirm & text
            update_changelog_on_bump = self._ask_update_changelog_on_bump()  # confirm
            major_version_zero = self._ask_major_version_zero(version)  # confirm
            hook_types: list[str] | None = questionary.checkbox(
                "What types of pre-commit hook you want to install? (Leave blank if you don't want to install)",
                choices=[
                    questionary.Choice("commit-msg", checked=False),
                    questionary.Choice("pre-push", checked=False),
                ],
            ).unsafe_ask()
        except KeyboardInterrupt:
            raise InitFailedError("Stopped by user")

        if hook_types:
            config_data = self._get_config_data()
            with smart_open(
                self._PRE_COMMIT_CONFIG_PATH,
                "w",
                encoding=self.config.settings["encoding"],
            ) as config_file:
                yaml.safe_dump(config_data, stream=config_file)

            if not project_info.is_pre_commit_installed():
                raise InitFailedError(
                    "Failed to install pre-commit hook.\n"
                    "pre-commit is not installed in current environment."
                )

            cmd_str = "pre-commit install " + " ".join(
                f"--hook-type {ty}" for ty in hook_types
            )
            c = cmd.run(cmd_str)
            if c.return_code != 0:
                raise InitFailedError(
                    "Failed to install pre-commit hook.\n"
                    f"Error running {cmd_str}."
                    "Outputs are attached below:\n"
                    f"stdout: {c.out}\n"
                    f"stderr: {c.err}"
                )
            out.write("commitizen pre-commit hook is now installed in your '.git'\n")

        _write_config_to_file(
            path=config_path,
            cz_name=cz_name,
            version_provider=version_provider,
            version_scheme=version_scheme,
            version=version,
            tag_format=tag_format,
            update_changelog_on_bump=update_changelog_on_bump,
            major_version_zero=major_version_zero,
        )

        out.write("\nYou can bump the version running:\n")
        out.info("\tcz bump\n")
        out.success("Configuration complete 🚀")

    def _ask_config_path(self) -> Path:
        filename: str = questionary.select(
            "Please choose a supported config file: ",
            choices=CONFIG_FILES,
            default=project_info.get_default_config_filename(),
            style=self.cz.style,
        ).unsafe_ask()
        return Path(filename)

    def _ask_name(self) -> str:
        name: str = questionary.select(
            "Please choose a cz (commit rule): (default: cz_conventional_commits)",
            choices=list(registry.keys()),
            default="cz_conventional_commits",
            style=self.cz.style,
        ).unsafe_ask()
        return name

    def _ask_tag(self) -> str:
        latest_tag = get_latest_tag_name()
        if not latest_tag:
            out.error("No Existing Tag. Set tag to v0.0.1")
            return "0.0.1"

        if questionary.confirm(
            f"Is {latest_tag} the latest tag?", style=self.cz.style, default=False
        ).unsafe_ask():
            return latest_tag

        existing_tags = get_tag_names()
        if not existing_tags:
            out.error("No Existing Tag. Set tag to v0.0.1")
            return "0.0.1"

        answer: str = questionary.select(
            "Please choose the latest tag: ",
            # The latest tag is most likely with the largest number.
            # Thus, listing the existing_tags in reverse order makes more sense.
            choices=sorted(existing_tags, reverse=True),
            style=self.cz.style,
        ).unsafe_ask()

        if not answer:
            raise NoAnswersError("Tag is required!")
        return answer

    def _ask_tag_format(self, latest_tag: str) -> str:
        if latest_tag.startswith("v"):
            v_tag_format = r"v$version"
            if questionary.confirm(
                f'Is "{v_tag_format}" the correct tag format?', style=self.cz.style
            ).unsafe_ask():
                return v_tag_format

        default_format = DEFAULT_SETTINGS["tag_format"]
        tag_format: str = questionary.text(
            f'Please enter the correct version format: (default: "{default_format}")',
            style=self.cz.style,
        ).unsafe_ask()

        return tag_format or default_format

    def _ask_version_provider(self) -> str:
        """Ask for setting: version_provider"""

        version_provider: str = questionary.select(
            "Choose the source of the version:",
            choices=_VERSION_PROVIDER_CHOICES,
            style=self.cz.style,
            default=project_info.get_default_version_provider(),
        ).unsafe_ask()
        return version_provider

    def _ask_version_scheme(self) -> str:
        """Ask for setting: version_scheme"""
        scheme: str = questionary.select(
            "Choose version scheme: ",
            choices=KNOWN_SCHEMES,
            style=self.cz.style,
            default=project_info.get_default_version_scheme(),
        ).unsafe_ask()
        return scheme

    def _ask_major_version_zero(self, version: Version) -> bool:
        """Ask for setting: major_version_zero"""
        if version.major > 0:
            return False
        major_version_zero: bool = questionary.confirm(
            "Keep major version zero (0.x) during breaking changes",
            default=True,
            auto_enter=True,
        ).unsafe_ask()
        return major_version_zero

    def _ask_update_changelog_on_bump(self) -> bool:
        """Ask for setting: update_changelog_on_bump"""
        update_changelog_on_bump: bool = questionary.confirm(
            "Create changelog automatically on bump",
            default=True,
            auto_enter=True,
        ).unsafe_ask()
        return update_changelog_on_bump

    def _get_config_data(self) -> dict[str, Any]:
        CZ_HOOK_CONFIG = {
            "repo": "https://github.com/commitizen-tools/commitizen",
            "rev": f"v{__version__}",
            "hooks": [
                {"id": "commitizen"},
                {"id": "commitizen-branch", "stages": ["pre-push"]},
            ],
        }

        if not Path(".pre-commit-config.yaml").is_file():
            return {"repos": [CZ_HOOK_CONFIG]}

        with open(
            self._PRE_COMMIT_CONFIG_PATH, encoding=self.config.settings["encoding"]
        ) as config_file:
            config_data: dict[str, Any] = yaml.safe_load(config_file) or {}

        if not isinstance(repos := config_data.get("repos"), list):
            # .pre-commit-config.yaml exists but there's no "repos" key
            config_data["repos"] = [CZ_HOOK_CONFIG]
            return config_data

        # Check if commitizen pre-commit hook is already in the config
        if any("commitizen" in hook_config["repo"] for hook_config in repos):
            out.write("commitizen already in pre-commit config")
        else:
            repos.append(CZ_HOOK_CONFIG)
        return config_data


def _write_config_to_file(
    *,
    path: Path,
    cz_name: str,
    version_provider: str,
    version_scheme: str,
    version: Version,
    tag_format: str,
    update_changelog_on_bump: bool,
    major_version_zero: bool,
) -> None:
    out_config = create_config(path=path)
    out_config.init_empty_config_content()

    out_config.set_key("name", cz_name)
    out_config.set_key("tag_format", tag_format)
    out_config.set_key("version_scheme", version_scheme)
    if version_provider == "commitizen":
        out_config.set_key("version", version.public)
    else:
        out_config.set_key("version_provider", version_provider)
    if update_changelog_on_bump:
        out_config.set_key("update_changelog_on_bump", update_changelog_on_bump)
    if major_version_zero:
        out_config.set_key("major_version_zero", major_version_zero)



"""Resolves project information about the current working directory."""

import shutil
from pathlib import Path
from typing import Literal


def is_pre_commit_installed() -> bool:
    return bool(shutil.which("pre-commit"))


def get_default_version_provider() -> Literal[
    "commitizen", "cargo", "composer", "npm", "pep621", "poetry", "uv"
]:
    pyproject_path = Path("pyproject.toml")
    if pyproject_path.is_file():
        if "[tool.poetry]" in pyproject_path.read_text():
            return "poetry"
        if Path("uv.lock").is_file():
            return "uv"
        return "pep621"

    if Path("setup.py").is_file():
        return "pep621"

    if Path("Cargo.toml").is_file():
        return "cargo"

    if Path("package.json").is_file():
        return "npm"

    if Path("composer.json").is_file():
        return "composer"

    return "commitizen"


def get_default_config_filename() -> Literal["pyproject.toml", ".cz.toml"]:
    return "pyproject.toml" if Path("pyproject.toml").is_file() else ".cz.toml"


def get_default_version_scheme() -> Literal["pep440", "semver"]:
    return (
        "pep440"
        if Path("pyproject.toml").is_file() or Path("setup.py").is_file()
        else "semver"
    )



from __future__ import annotations

from commitizen.git import get_tags
from commitizen.providers.base_provider import VersionProvider
from commitizen.tags import TagRules


class ScmProvider(VersionProvider):
    """
    A provider fetching the current/last version from the repository history

    The version is fetched using `git describe` and is never set.

    It is meant for `setuptools-scm` or any package manager `*-scm` provider.
    """

    def get_version(self) -> str:
        rules = TagRules.from_settings(self.config.settings)
        tags = get_tags(reachable_only=True)
        version_tags = rules.get_version_tags(tags)
        version = max((rules.extract_version(t) for t in version_tags), default=None)
        return str(version) if version is not None else "0.0.0"

    def set_version(self, version: str) -> None:
        # Not necessary
        pass



from __future__ import annotations

from textwrap import dedent
from typing import TYPE_CHECKING

import pytest

from commitizen.providers import get_provider
from commitizen.providers.poetry_provider import PoetryProvider

if TYPE_CHECKING:
    from pathlib import Path

    from commitizen.config.base_config import BaseConfig

POETRY_TOML = """\
[tool.poetry]
version = "0.1.0"
"""

POETRY_EXPECTED = """\
[tool.poetry]
version = "42.1"
"""


@pytest.mark.parametrize(
    "content, expected",
    ((POETRY_TOML, POETRY_EXPECTED),),
)
def test_cargo_provider(
    config: BaseConfig,
    chdir: Path,
    content: str,
    expected: str,
):
    filename = PoetryProvider.filename
    file = chdir / filename
    file.write_text(dedent(content))
    config.settings["version_provider"] = "poetry"

    provider = get_provider(config)
    assert isinstance(provider, PoetryProvider)
    assert provider.get_version() == "0.1.0"

    provider.set_version("42.1")
    assert file.read_text() == dedent(expected)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from commitizen.exceptions import VersionProviderUnknown
from commitizen.providers import get_provider
from commitizen.providers.commitizen_provider import CommitizenProvider

if TYPE_CHECKING:
    from commitizen.config.base_config import BaseConfig


def test_default_version_provider_is_commitizen_config(config: BaseConfig):
    provider = get_provider(config)

    assert isinstance(provider, CommitizenProvider)


def test_raise_for_unknown_provider(config: BaseConfig):
    config.settings["version_provider"] = "unknown"
    with pytest.raises(VersionProviderUnknown):
        get_provider(config)



import os
import subprocess
import sys
import types
from functools import partial

import pytest
from pytest_mock import MockFixture

from commitizen import cli
from commitizen.exceptions import (
    ConfigFileNotFound,
    ExpectedExit,
    InvalidCommandArgumentError,
    NoCommandFoundError,
    NotAGitProjectError,
)


def test_sysexit_no_argv(mocker: MockFixture, capsys):
    testargs = ["cz"]
    mocker.patch.object(sys, "argv", testargs)

    with pytest.raises(ExpectedExit):
        cli.main()
        out, _ = capsys.readouterr()
        assert out.startswith("usage")


def test_cz_config_file_without_correct_file_path(mocker: MockFixture, capsys):
    testargs = ["cz", "--config", "./config/pyproject.toml", "example"]
    mocker.patch.object(sys, "argv", testargs)

    with pytest.raises(ConfigFileNotFound) as excinfo:
        cli.main()
    assert "Cannot found the config file" in str(excinfo.value)


def test_cz_with_arg_but_without_command(mocker: MockFixture):
    testargs = ["cz", "--name", "cz_jira"]
    mocker.patch.object(sys, "argv", testargs)

    with pytest.raises(NoCommandFoundError) as excinfo:
        cli.main()
    assert "Command is required" in str(excinfo.value)


def test_name(mocker: MockFixture, capsys):
    testargs = ["cz", "-n", "cz_jira", "example"]
    mocker.patch.object(sys, "argv", testargs)

    cli.main()
    out, _ = capsys.readouterr()
    assert out.startswith("JRA")


@pytest.mark.usefixtures("tmp_git_project")
def test_name_default_value(mocker: MockFixture, capsys):
    testargs = ["cz", "example"]
    mocker.patch.object(sys, "argv", testargs)

    cli.main()
    out, _ = capsys.readouterr()
    assert out.startswith("fix: correct minor typos in code")


def test_ls(mocker: MockFixture, capsys):
    testargs = ["cz", "-n", "cz_jira", "ls"]
    mocker.patch.object(sys, "argv", testargs)
    cli.main()
    out, err = capsys.readouterr()

    assert "cz_conventional_commits" in out
    assert isinstance(out, str)


def test_arg_debug(mocker: MockFixture):
    testargs = ["cz", "--debug", "info"]
    mocker.patch.object(sys, "argv", testargs)
    cli.main()
    excepthook = sys.excepthook
    # `sys.excepthook` is replaced by a `partial` in `cli.main`
    # it's not a normal function
    assert isinstance(excepthook, partial)
    assert excepthook.keywords.get("debug") is True


def test_commitizen_excepthook(capsys):
    with pytest.raises(SystemExit) as excinfo:
        cli.commitizen_excepthook(NotAGitProjectError, NotAGitProjectError(), "")

    assert excinfo.type is SystemExit
    assert excinfo.value.code == NotAGitProjectError.exit_code


def test_commitizen_debug_excepthook(capsys):
    with pytest.raises(SystemExit) as excinfo:
        cli.commitizen_excepthook(
            NotAGitProjectError,
            NotAGitProjectError(),
            "",
            debug=True,
        )

    assert excinfo.type is SystemExit
    assert excinfo.value.code == NotAGitProjectError.exit_code
    assert "NotAGitProjectError" in str(excinfo.traceback[0])


@pytest.mark.skipif(
    os.name == "nt",
    reason="`argcomplete` does not support Git Bash on Windows.",
)
def test_argcomplete_activation():
    """
    This function is testing the one-time activation of argcomplete for
    commitizen only.

    Equivalent to run:
    $ eval "$(register-python-argcomplete pytest)"
    """
    output = subprocess.run(["register-python-argcomplete", "cz"])

    assert output.returncode == 0


def test_commitizen_excepthook_no_raises(capsys):
    with pytest.raises(SystemExit) as excinfo:
        cli.commitizen_excepthook(
            NotAGitProjectError,
            NotAGitProjectError(),
            "",
            no_raise=[NotAGitProjectError.exit_code],
        )

    assert excinfo.type is SystemExit
    assert excinfo.value.code == 0


def test_parse_no_raise_single_integer():
    input_str = "1"
    result = cli.parse_no_raise(input_str)
    assert result == [1]


def test_parse_no_raise_integers():
    input_str = "1,2,3"
    result = cli.parse_no_raise(input_str)
    assert result == [1, 2, 3]


def test_parse_no_raise_error_code():
    input_str = "NO_COMMITIZEN_FOUND,NO_COMMITS_FOUND,NO_PATTERN_MAP"
    result = cli.parse_no_raise(input_str)
    assert result == [1, 3, 5]


def test_parse_no_raise_mix_integer_error_code():
    input_str = "NO_COMMITIZEN_FOUND,2,NO_COMMITS_FOUND,4"
    result = cli.parse_no_raise(input_str)
    assert result == [1, 2, 3, 4]


def test_parse_no_raise_mix_invalid_arg_is_skipped():
    input_str = "NO_COMMITIZEN_FOUND,2,nothing,4"
    result = cli.parse_no_raise(input_str)
    assert result == [1, 2, 4]


def test_unknown_args_raises(mocker: MockFixture):
    testargs = ["cz", "c", "-this_arg_is_not_supported"]
    mocker.patch.object(sys, "argv", testargs)
    with pytest.raises(InvalidCommandArgumentError) as excinfo:
        cli.main()
    assert "Invalid commitizen arguments were found" in str(excinfo.value)


def test_unknown_args_before_double_dash_raises(mocker: MockFixture):
    testargs = ["cz", "c", "-this_arg_is_not_supported", "--"]
    mocker.patch.object(sys, "argv", testargs)
    with pytest.raises(InvalidCommandArgumentError) as excinfo:
        cli.main()
    assert "Invalid commitizen arguments were found before -- separator" in str(
        excinfo.value
    )


def test_commitizen_excepthook_non_commitizen_exception(mocker: MockFixture):
    """Test that commitizen_excepthook delegates to sys.__excepthook__ for non-CommitizenException."""
    # Mock the original excepthook
    mock_original_excepthook = mocker.Mock()
    mocker.patch("commitizen.cli.sys.__excepthook__", mock_original_excepthook)

    # Create a regular exception
    test_exception = ValueError("test error")

    # Call commitizen_excepthook with the regular exception
    cli.commitizen_excepthook(ValueError, test_exception, None)

    # Verify sys.__excepthook__ was called with correct arguments
    mock_original_excepthook.assert_called_once_with(ValueError, test_exception, None)


def test_commitizen_excepthook_non_commitizen_exception_with_traceback(
    mocker: MockFixture,
):
    """Test that commitizen_excepthook handles traceback correctly for non-CommitizenException."""
    # Mock the original excepthook
    mock_original_excepthook = mocker.Mock()
    mocker.patch("commitizen.cli.sys.__excepthook__", mock_original_excepthook)

    # Create a regular exception with a traceback
    test_exception = ValueError("test error")
    test_traceback = mocker.Mock(spec=types.TracebackType)

    # Call commitizen_excepthook with the regular exception and traceback
    cli.commitizen_excepthook(ValueError, test_exception, test_traceback)

    # Verify sys.__excepthook__ was called with correct arguments including traceback
    mock_original_excepthook.assert_called_once_with(
        ValueError, test_exception, test_traceback
    )


def test_commitizen_excepthook_non_commitizen_exception_with_invalid_traceback(
    mocker: MockFixture,
):
    """Test that commitizen_excepthook handles invalid traceback correctly for non-CommitizenException."""
    # Mock the original excepthook
    mock_original_excepthook = mocker.Mock()
    mocker.patch("commitizen.cli.sys.__excepthook__", mock_original_excepthook)

    # Create a regular exception with an invalid traceback
    test_exception = ValueError("test error")
    test_traceback = mocker.Mock()  # Not a TracebackType

    # Call commitizen_excepthook with the regular exception and invalid traceback
    cli.commitizen_excepthook(ValueError, test_exception, test_traceback)

    # Verify sys.__excepthook__ was called with None as traceback
    mock_original_excepthook.assert_called_once_with(ValueError, test_exception, None)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Mapping

    from jinja2 import Template

    from commitizen.config import BaseConfig
    from commitizen.question import CzQuestion
else:
    try:
        from jinja2 import Template
    except ImportError:
        from string import Template


from commitizen import defaults
from commitizen.cz.base import BaseCommitizen
from commitizen.exceptions import MissingCzCustomizeConfigError

__all__ = ["CustomizeCommitsCz"]


class CustomizeCommitsCz(BaseCommitizen):
    bump_pattern = defaults.BUMP_PATTERN
    bump_map = defaults.BUMP_MAP
    bump_map_major_version_zero = defaults.BUMP_MAP_MAJOR_VERSION_ZERO
    change_type_order = defaults.CHANGE_TYPE_ORDER

    def __init__(self, config: BaseConfig) -> None:
        super().__init__(config)

        if "customize" not in self.config.settings:
            raise MissingCzCustomizeConfigError()
        self.custom_settings = self.config.settings["customize"]

        for attr_name in [
            "bump_pattern",
            "bump_map",
            "bump_map_major_version_zero",
            "change_type_order",
            "commit_parser",
            "changelog_pattern",
            "change_type_map",
        ]:
            if value := self.custom_settings.get(attr_name):
                setattr(self, attr_name, value)

    def questions(self) -> list[CzQuestion]:
        return self.custom_settings.get("questions", [{}])  # type: ignore[return-value]

    def message(self, answers: Mapping[str, Any]) -> str:
        message_template = Template(self.custom_settings.get("message_template", ""))
        if getattr(Template, "substitute", None):
            return message_template.substitute(**answers)  # type: ignore[attr-defined,no-any-return] # pragma: no cover # TODO: check if we can fix this
        return message_template.render(**answers)

    def example(self) -> str:
        return self.custom_settings.get("example") or ""

    def schema_pattern(self) -> str:
        return self.custom_settings.get("schema_pattern") or ""

    def schema(self) -> str:
        return self.custom_settings.get("schema") or ""

    def info(self) -> str:
        if info_path := self.custom_settings.get("info_path"):
            with open(info_path, encoding=self.config.settings["encoding"]) as f:
                return f.read()
        return self.custom_settings.get("info") or ""



import pytest

from commitizen import changelog_formats, defaults


def test_getattr_deprecated_vars():
    # Test each deprecated variable
    with pytest.warns(DeprecationWarning) as record:
        assert defaults.bump_pattern == defaults.BUMP_PATTERN
        assert defaults.bump_map == defaults.BUMP_MAP
        assert (
            defaults.bump_map_major_version_zero == defaults.BUMP_MAP_MAJOR_VERSION_ZERO
        )
        assert defaults.bump_message == defaults.BUMP_MESSAGE
        assert defaults.change_type_order == defaults.CHANGE_TYPE_ORDER
        assert defaults.encoding == defaults.ENCODING
        assert defaults.name == defaults.DEFAULT_SETTINGS["name"]
        assert (
            changelog_formats._guess_changelog_format
            == changelog_formats.guess_changelog_format
        )

    # Verify warning messages
    assert len(record) == 7
    for warning in record:
        assert "is deprecated and will be removed" in str(warning.message)


def test_getattr_non_existent():
    # Test non-existent attribute
    with pytest.raises(AttributeError) as exc_info:
        _ = defaults.non_existent_attribute
    assert "is not an attribute of" in str(exc_info.value)



from __future__ import annotations

import fnmatch
import glob
from pathlib import Path
from typing import TYPE_CHECKING

from tomlkit import TOMLDocument, dumps, parse
from tomlkit.exceptions import NonExistentKey

from commitizen.providers.base_provider import TomlProvider

if TYPE_CHECKING:
    from tomlkit.items import AoT


class CargoProvider(TomlProvider):
    """
    Cargo version management

    With support for `workspaces`
    """

    filename = "Cargo.toml"
    lock_filename = "Cargo.lock"

    @property
    def lock_file(self) -> Path:
        return Path() / self.lock_filename

    def get(self, document: TOMLDocument) -> str:
        out = _try_get_workspace(document)["package"]["version"]
        if TYPE_CHECKING:
            assert isinstance(out, str)
        return out

    def set(self, document: TOMLDocument, version: str) -> None:
        _try_get_workspace(document)["package"]["version"] = version

    def set_version(self, version: str) -> None:
        super().set_version(version)
        if self.lock_file.exists():
            self.set_lock_version(version)

    def set_lock_version(self, version: str) -> None:
        cargo_toml_content = parse(self.file.read_text())
        cargo_lock_content = parse(self.lock_file.read_text())
        packages = cargo_lock_content["package"]

        if TYPE_CHECKING:
            assert isinstance(packages, AoT)

        try:
            cargo_package_name = cargo_toml_content["package"]["name"]  # type: ignore[index]
            if TYPE_CHECKING:
                assert isinstance(cargo_package_name, str)
            for i, package in enumerate(packages):
                if package["name"] == cargo_package_name:
                    cargo_lock_content["package"][i]["version"] = version  # type: ignore[index]
                    break
        except NonExistentKey:
            workspace = cargo_toml_content.get("workspace", {})
            if TYPE_CHECKING:
                assert isinstance(workspace, dict)
            workspace_members = workspace.get("members", [])
            excluded_workspace_members = workspace.get("exclude", [])
            members_inheriting: list[str] = []

            for member in workspace_members:
                for path in glob.glob(member, recursive=True):
                    if any(
                        fnmatch.fnmatch(path, pattern)
                        for pattern in excluded_workspace_members
                    ):
                        continue

                    cargo_file = Path(path) / "Cargo.toml"
                    package_content = parse(cargo_file.read_text()).get("package", {})
                    if TYPE_CHECKING:
                        assert isinstance(package_content, dict)
                    try:
                        version_workspace = package_content["version"]["workspace"]
                        if version_workspace is True:
                            package_name = package_content["name"]
                            if TYPE_CHECKING:
                                assert isinstance(package_name, str)
                            members_inheriting.append(package_name)
                    except NonExistentKey:
                        pass

            for i, package in enumerate(packages):
                if package["name"] in members_inheriting:
                    cargo_lock_content["package"][i]["version"] = version  # type: ignore[index]

        self.lock_file.write_text(dumps(cargo_lock_content))


def _try_get_workspace(document: TOMLDocument) -> dict:
    try:
        workspace = document["workspace"]
        if TYPE_CHECKING:
            assert isinstance(workspace, dict)
        return workspace
    except NonExistentKey:
        return document



from __future__ import annotations

import os
from enum import Enum
from functools import lru_cache
from pathlib import Path
from tempfile import NamedTemporaryFile

from commitizen import cmd, out
from commitizen.exceptions import GitCommandError


class EOLType(Enum):
    """The EOL type from `git config core.eol`."""

    LF = "lf"
    CRLF = "crlf"
    NATIVE = "native"

    @classmethod
    def for_open(cls) -> str:
        c = cmd.run("git config core.eol")
        eol = c.out.strip().upper()
        return cls._char_for_open()[cls._safe_cast(eol)]

    @classmethod
    def _safe_cast(cls, eol: str) -> EOLType:
        try:
            return cls[eol]
        except KeyError:
            return cls.NATIVE

    @classmethod
    @lru_cache
    def _char_for_open(cls) -> dict[EOLType, str]:
        """Get the EOL character for `open()`."""
        return {
            cls.LF: "\n",
            cls.CRLF: "\r\n",
            cls.NATIVE: os.linesep,
        }


class GitObject:
    rev: str
    name: str
    date: str

    def __eq__(self, other: object) -> bool:
        return hasattr(other, "rev") and self.rev == other.rev

    def __hash__(self) -> int:
        return hash(self.rev)


class GitCommit(GitObject):
    def __init__(
        self,
        rev: str,
        title: str,
        body: str = "",
        author: str = "",
        author_email: str = "",
        parents: list[str] | None = None,
    ) -> None:
        self.rev = rev.strip()
        self.title = title.strip()
        self.body = body.strip()
        self.author = author.strip()
        self.author_email = author_email.strip()
        self.parents = parents or []

    @property
    def message(self) -> str:
        return f"{self.title}\n\n{self.body}".strip()

    @classmethod
    def from_rev_and_commit(cls, rev_and_commit: str) -> GitCommit:
        """Create a GitCommit instance from a formatted commit string.

        This method parses a multi-line string containing commit information in the following format:
        ```
        <rev>
        <parents>
        <title>
        <author>
        <author_email>
        <body_line_1>
        <body_line_2>
        ...
        ```

        Args:
            rev_and_commit (str): A string containing commit information with fields separated by newlines.
                - rev: The commit hash/revision
                - parents: Space-separated list of parent commit hashes
                - title: The commit title/message
                - author: The commit author's name
                - author_email: The commit author's email
                - body: Optional multi-line commit body

        Returns:
            GitCommit: A new GitCommit instance with the parsed information.

        Example:
            >>> commit_str = '''abc123
            ... def456 ghi789
            ... feat: add new feature
            ... John Doe
            ... john@example.com
            ... This is a detailed description
            ... of the new feature'''
            >>> commit = GitCommit.from_rev_and_commit(commit_str)
            >>> commit.rev
            'abc123'
            >>> commit.title
            'feat: add new feature'
            >>> commit.parents
            ['def456', 'ghi789']
        """
        rev, parents, title, author, author_email, *body_list = rev_and_commit.split(
            "\n"
        )
        return cls(
            rev=rev.strip(),
            title=title.strip(),
            body="\n".join(body_list).strip(),
            author=author,
            author_email=author_email,
            parents=[p for p in parents.strip().split(" ") if p],
        )

    def __repr__(self) -> str:
        return f"{self.title} ({self.rev})"


class GitTag(GitObject):
    def __init__(self, name: str, rev: str, date: str) -> None:
        self.rev = rev.strip()
        self.name = name.strip()
        self._date = date.strip()

    def __repr__(self) -> str:
        return f"GitTag('{self.name}', '{self.rev}', '{self.date}')"

    @property
    def date(self) -> str:
        return self._date

    @date.setter
    def date(self, value: str) -> None:
        self._date = value

    @classmethod
    def from_line(cls, line: str, inner_delimiter: str) -> GitTag:
        name, objectname, date, obj = line.split(inner_delimiter)
        if not obj:
            obj = objectname

        return cls(name=name, rev=obj, date=date)


def tag(
    tag: str, annotated: bool = False, signed: bool = False, msg: str | None = None
) -> cmd.Command:
    if not annotated and not signed:
        return cmd.run(f"git tag {tag}")

    # according to https://git-scm.com/book/en/v2/Git-Basics-Tagging,
    # we're not able to create lightweight tag with message.
    # by adding message, we make it a annotated tags
    option = "-s" if signed else "-a"  # The else case is for annotated tags
    return cmd.run(f'git tag {option} {tag} -m "{msg or tag}"')


def add(*args: str) -> cmd.Command:
    return cmd.run(f"git add {' '.join(args)}")


def commit(
    message: str,
    args: str = "",
    committer_date: str | None = None,
) -> cmd.Command:
    f = NamedTemporaryFile("wb", delete=False)
    f.write(message.encode("utf-8"))
    f.close()

    command = _create_commit_cmd_string(args, committer_date, f.name)
    c = cmd.run(command)
    os.unlink(f.name)
    return c


def _create_commit_cmd_string(args: str, committer_date: str | None, name: str) -> str:
    command = f'git commit {args} -F "{name}"'
    if not committer_date:
        return command
    if os.name != "nt":
        return f"GIT_COMMITTER_DATE={committer_date} {command}"
    # Using `cmd /v /c "{command}"` sets environment variables only for that command
    return f'cmd /v /c "set GIT_COMMITTER_DATE={committer_date}&& {command}"'


def get_commits(
    start: str | None = None,
    end: str | None = None,
    *,
    args: str = "",
) -> list[GitCommit]:
    """Get the commits between start and end."""
    if end is None:
        end = "HEAD"
    git_log_entries = _get_log_as_str_list(start, end, args)
    return [
        GitCommit.from_rev_and_commit(rev_and_commit)
        for rev_and_commit in git_log_entries
        if rev_and_commit
    ]


def get_filenames_in_commit(git_reference: str = "") -> list[str]:
    """Get the list of files that were committed in the requested git reference.

    :param git_reference: a git reference as accepted by `git show`, default: the last commit

    :returns: file names committed in the last commit by default or inside the passed git reference
    """
    c = cmd.run(f"git show --name-only --pretty=format: {git_reference}")
    if c.return_code == 0:
        return c.out.strip().split("\n")
    raise GitCommandError(c.err)


def get_tags(
    dateformat: str = "%Y-%m-%d", reachable_only: bool = False
) -> list[GitTag]:
    inner_delimiter = "---inner_delimiter---"
    formatter = (
        f'"%(refname:strip=2){inner_delimiter}'
        f"%(objectname){inner_delimiter}"
        f"%(creatordate:format:{dateformat}){inner_delimiter}"
        f'%(object)"'
    )
    extra = "--merged" if reachable_only else ""
    # Force the default language for parsing
    env = {"LC_ALL": "C", "LANG": "C", "LANGUAGE": "C"}
    c = cmd.run(f"git tag --format={formatter} --sort=-creatordate {extra}", env=env)
    if c.return_code != 0:
        if reachable_only and c.err == "fatal: malformed object name HEAD\n":
            # this can happen if there are no commits in the repo yet
            return []
        raise GitCommandError(c.err)

    if c.err:
        out.warn(f"Attempting to proceed after: {c.err}")

    return [
        GitTag.from_line(line=line, inner_delimiter=inner_delimiter)
        for line in c.out.split("\n")[:-1]
    ]


def tag_exist(tag: str) -> bool:
    c = cmd.run(f"git tag --list {tag}")
    return tag in c.out


def is_signed_tag(tag: str) -> bool:
    return cmd.run(f"git tag -v {tag}").return_code == 0


def get_latest_tag_name() -> str | None:
    c = cmd.run("git describe --abbrev=0 --tags")
    if c.err:
        return None
    return c.out.strip()


def get_tag_message(tag: str) -> str | None:
    c = cmd.run(f"git tag -l --format='%(contents:subject)' {tag}")
    if c.err:
        return None
    return c.out.strip()


def get_tag_names() -> list[str]:
    c = cmd.run("git tag --list")
    if c.err:
        return []
    return [tag for raw in c.out.split("\n") if (tag := raw.strip())]


def find_git_project_root() -> Path | None:
    c = cmd.run("git rev-parse --show-toplevel")
    if c.err:
        return None
    return Path(c.out.strip())


def is_staging_clean() -> bool:
    """Check if staging is clean."""
    c = cmd.run("git diff --no-ext-diff --cached --name-only")
    return not bool(c.out)


def is_git_project() -> bool:
    c = cmd.run("git rev-parse --is-inside-work-tree")
    return c.out.strip() == "true"


def get_core_editor() -> str | None:
    c = cmd.run("git var GIT_EDITOR")
    if c.out:
        return c.out.strip()
    return None


def smart_open(*args, **kwargs):  # type: ignore[no-untyped-def,unused-ignore] # noqa: ANN201
    """Open a file with the EOL style determined from Git."""
    return open(*args, newline=EOLType.for_open(), **kwargs)


def _get_log_as_str_list(start: str | None, end: str, args: str) -> list[str]:
    """Get string representation of each log entry"""
    delimiter = "----------commit-delimiter----------"
    log_format: str = "%H%n%P%n%s%n%an%n%ae%n%b"
    command_range = f"{start}..{end}" if start else end
    command = f"git -c log.showSignature=False log --pretty={log_format}{delimiter} {args} {command_range}"

    c = cmd.run(command)
    if c.return_code != 0:
        raise GitCommandError(c.err)
    return c.out.split(f"{delimiter}\n")


def get_default_branch() -> str:
    c = cmd.run("git symbolic-ref refs/remotes/origin/HEAD")
    if c.return_code != 0:
        raise GitCommandError(c.err)
    return c.out.strip()



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from commitizen.providers import get_provider
from commitizen.providers.scm_provider import ScmProvider
from tests.utils import (
    create_branch,
    create_file_and_commit,
    create_tag,
    merge_branch,
    switch_branch,
)

if TYPE_CHECKING:
    from commitizen.config.base_config import BaseConfig


@pytest.mark.parametrize(
    "tag_format,tag,expected_version",
    (
        # If tag_format is $version (the default), version_scheme.parser is used.
        # Its DEFAULT_VERSION_PARSER allows a v prefix, but matches PEP440 otherwise.
        ("$version", "no-match-because-version-scheme-is-strict", "0.0.0"),
        ("$version", "0.1.0", "0.1.0"),
        ("$version", "v0.1.0", "0.1.0"),
        ("$version", "v-0.1.0", "0.0.0"),
        # If tag_format is not None or $version, TAG_FORMAT_REGEXS are used, which are
        # much more lenient but require a v prefix.
        ("v$version", "v0.1.0", "0.1.0"),
        ("v$version", "no-match-because-no-v-prefix", "0.0.0"),
        # no match because not a valid version
        ("v$version", "v-match-TAG_FORMAT_REGEXS", "0.0.0"),
        ("version-$version", "version-0.1.0", "0.1.0"),
        ("version-$version", "version-0.1", "0.1"),
        ("version-$version", "version-0.1.0rc1", "0.1.0rc1"),
        ("v$minor.$major.$patch", "v1.0.0", "0.1.0"),
        ("version-$major.$minor.$patch", "version-0.1.0", "0.1.0"),
        ("v$major.$minor$prerelease$devrelease", "v1.0rc1", "1.0rc1"),
        ("v$major.$minor.$patch$prerelease$devrelease", "v0.1.0", "0.1.0"),
        ("v$major.$minor.$patch$prerelease$devrelease", "v0.1.0rc1", "0.1.0rc1"),
        ("v$major.$minor.$patch$prerelease$devrelease", "v1.0.0.dev0", "1.0.0.dev0"),
    ),
)
@pytest.mark.usefixtures("tmp_git_project")
def test_scm_provider(
    config: BaseConfig, tag_format: str, tag: str, expected_version: str
):
    create_file_and_commit("test: fake commit")
    create_tag(tag)
    create_file_and_commit("test: fake commit")
    create_tag("should-not-match")

    config.settings["version_provider"] = "scm"
    config.settings["tag_format"] = tag_format

    provider = get_provider(config)
    assert isinstance(provider, ScmProvider)
    actual_version = provider.get_version()
    assert actual_version == expected_version

    # Should not fail on set_version()
    provider.set_version("43.1")


@pytest.mark.usefixtures("tmp_git_project")
def test_scm_provider_default_without_commits_and_tags(config: BaseConfig):
    config.settings["version_provider"] = "scm"

    provider = get_provider(config)
    assert isinstance(provider, ScmProvider)
    assert provider.get_version() == "0.0.0"


@pytest.mark.usefixtures("tmp_git_project")
def test_scm_provider_default_with_commits_and_tags(config: BaseConfig):
    config.settings["version_provider"] = "scm"

    provider = get_provider(config)
    assert isinstance(provider, ScmProvider)
    assert provider.get_version() == "0.0.0"

    create_file_and_commit("Initial state")
    create_tag("1.0.0")
    # create develop
    create_branch("develop")
    switch_branch("develop")

    # add a feature to develop
    create_file_and_commit("develop: add beta feature1")
    assert provider.get_version() == "1.0.0"
    create_tag("1.1.0b0")

    # create staging
    create_branch("staging")
    switch_branch("staging")
    create_file_and_commit("staging: Starting release candidate")
    assert provider.get_version() == "1.1.0b0"
    create_tag("1.1.0rc0")

    # add another feature to develop
    switch_branch("develop")
    create_file_and_commit("develop: add beta feature2")
    assert provider.get_version() == "1.1.0b0"
    create_tag("1.2.0b0")

    # add a hotfix to master
    switch_branch("master")
    create_file_and_commit("master: add hotfix")
    assert provider.get_version() == "1.0.0"
    create_tag("1.0.1")

    # merge the hotfix to staging
    switch_branch("staging")
    merge_branch("master")

    assert provider.get_version() == "1.1.0rc0"


@pytest.mark.usefixtures("tmp_git_project")
def test_scm_provider_detect_legacy_tags(config: BaseConfig):
    config.settings["version_provider"] = "scm"
    config.settings["tag_format"] = "v${version}"
    config.settings["legacy_tag_formats"] = [
        "legacy-${version}",
        "old-${version}",
    ]
    provider = get_provider(config)

    create_file_and_commit("test: fake commit")
    create_tag("old-0.4.1")
    assert provider.get_version() == "0.4.1"

    create_file_and_commit("test: fake commit")
    create_tag("legacy-0.4.2")
    assert provider.get_version() == "0.4.2"

    create_file_and_commit("test: fake commit")
    create_tag("v0.5.0")
    assert provider.get_version() == "0.5.0"



from __future__ import annotations

from typing import TYPE_CHECKING

from commitizen.providers.base_provider import TomlProvider

if TYPE_CHECKING:
    import tomlkit


class PoetryProvider(TomlProvider):
    """
    Poetry version management
    """

    filename = "pyproject.toml"

    def get(self, pyproject: tomlkit.TOMLDocument) -> str:
        return pyproject["tool"]["poetry"]["version"]  # type: ignore[index,return-value]

    def set(self, pyproject: tomlkit.TOMLDocument, version: str) -> None:
        pyproject["tool"]["poetry"]["version"] = version  # type: ignore[index]



from __future__ import annotations

import json
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar

from commitizen.providers.base_provider import VersionProvider

if TYPE_CHECKING:
    from collections.abc import Mapping


class NpmProvider(VersionProvider):
    """
    npm package.json and package-lock.json version management
    """

    indent: ClassVar[int] = 2
    package_filename = "package.json"
    lock_filename = "package-lock.json"
    shrinkwrap_filename = "npm-shrinkwrap.json"

    @property
    def package_file(self) -> Path:
        return Path() / self.package_filename

    @property
    def lock_file(self) -> Path:
        return Path() / self.lock_filename

    @property
    def shrinkwrap_file(self) -> Path:
        return Path() / self.shrinkwrap_filename

    def get_version(self) -> str:
        """
        Get the current version from package.json
        """
        package_document = json.loads(self.package_file.read_text())
        return self.get_package_version(package_document)

    def set_version(self, version: str) -> None:
        package_document = self.set_package_version(
            json.loads(self.package_file.read_text()), version
        )
        self.package_file.write_text(
            json.dumps(package_document, indent=self.indent) + "\n"
        )
        if self.lock_file.exists():
            lock_document = self.set_lock_version(
                json.loads(self.lock_file.read_text()), version
            )
            self.lock_file.write_text(
                json.dumps(lock_document, indent=self.indent) + "\n"
            )
        if self.shrinkwrap_file.exists():
            shrinkwrap_document = self.set_shrinkwrap_version(
                json.loads(self.shrinkwrap_file.read_text()), version
            )
            self.shrinkwrap_file.write_text(
                json.dumps(shrinkwrap_document, indent=self.indent) + "\n"
            )

    def get_package_version(self, document: Mapping[str, str]) -> str:
        return document["version"]

    def set_package_version(
        self, document: dict[str, Any], version: str
    ) -> dict[str, Any]:
        document["version"] = version
        return document

    def set_lock_version(
        self, document: dict[str, Any], version: str
    ) -> dict[str, Any]:
        document["version"] = version
        document["packages"][""]["version"] = version
        return document

    def set_shrinkwrap_version(
        self, document: dict[str, Any], version: str
    ) -> dict[str, Any]:
        document["version"] = version
        document["packages"][""]["version"] = version
        return document



from __future__ import annotations

import os
import subprocess
from typing import TYPE_CHECKING, NamedTuple

from charset_normalizer import from_bytes

from commitizen.exceptions import CharacterSetDecodeError

if TYPE_CHECKING:
    from collections.abc import Mapping


class Command(NamedTuple):
    out: str
    err: str
    stdout: bytes
    stderr: bytes
    return_code: int


def _try_decode(bytes_: bytes) -> str:
    try:
        return bytes_.decode("utf-8")
    except UnicodeDecodeError:
        pass

    charset_match = from_bytes(bytes_).best()
    if charset_match is None:
        raise CharacterSetDecodeError()
    try:
        return bytes_.decode(charset_match.encoding)
    except UnicodeDecodeError as e:
        raise CharacterSetDecodeError() from e


def run(cmd: str, env: Mapping[str, str] | None = None) -> Command:
    if env is not None:
        env = {**os.environ, **env}
    process = subprocess.Popen(
        cmd,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE,
        env=env,
    )
    stdout, stderr = process.communicate()
    return_code = process.returncode
    return Command(
        _try_decode(stdout),
        _try_decode(stderr),
        stdout,
        stderr,
        return_code,
    )



import sys

import pytest
from pytest_mock import MockFixture

from commitizen import cli
from commitizen.commands import Example, Info, ListCz, Schema


@pytest.mark.skipif(
    sys.version_info < (3, 13),
    reason="The output message of argparse is different between Python 3.13 and lower than Python 3.13",
)
@pytest.mark.parametrize(
    "command",
    [
        "bump",
        "changelog",
        "check",
        "commit",
        "example",
        "info",
        "init",
        "ls",
        "schema",
        "version",
    ],
)
def test_command_shows_description_when_use_help_option(
    mocker: MockFixture,
    capsys,
    file_regression,
    monkeypatch: pytest.MonkeyPatch,
    command: str,
):
    """Test that the command shows the description when the help option is used.

    Note: If the command description changes, please run `pytest tests/commands/test_common_command.py --regen-all` to regenerate the test files.
    """
    # Force consistent terminal output
    monkeypatch.setenv("COLUMNS", "80")
    monkeypatch.setenv("TERM", "dumb")
    monkeypatch.setenv("LC_ALL", "C")
    monkeypatch.setenv("LANG", "C")
    monkeypatch.setenv("NO_COLOR", "1")
    monkeypatch.setenv("PAGER", "cat")

    testargs = ["cz", command, "--help"]
    mocker.patch.object(sys, "argv", testargs)
    with pytest.raises(SystemExit):
        cli.main()

    out, _ = capsys.readouterr()
    file_regression.check(out, extension=".txt")


@pytest.mark.parametrize(
    "command",
    [
        Example,
        Info,
        ListCz,
        Schema,
    ],
)
def test_simple_command_call_once(config, mocker: MockFixture, command):
    write_mock = mocker.patch("commitizen.out.write")
    command(config)()
    write_mock.assert_called_once()



import pytest

from commitizen import cmd
from commitizen.exceptions import CharacterSetDecodeError


# https://docs.python.org/3/howto/unicode.html
def test_valid_utf8_encoded_strings():
    valid_strings = (
        "",
        "ascii",
        "🤦🏻‍♂️",
        "﷽",
        "\u0000",
    )
    assert all(s == cmd._try_decode(s.encode("utf-8")) for s in valid_strings)


# A word of caution: just because an encoding can be guessed for a given
# sequence of bytes and because that guessed encoding may yield a decoded
# string, does not mean that that string was the original! For more, see:
# https://docs.python.org/3/library/codecs.html#standard-encodings


# Pick a random, non-utf8 encoding to test.
def test_valid_cp1250_encoded_strings():
    valid_strings = (
        "",
        "ascii",
        "äöüß",
        "ça va",
        "jak se máte",
    )
    for s in valid_strings:
        assert cmd._try_decode(s.encode("cp1250")) or True


def test_invalid_bytes():
    invalid_bytes = (b"\x73\xe2\x9d\xff\x00",)
    for s in invalid_bytes:
        with pytest.raises(CharacterSetDecodeError):
            cmd._try_decode(s)


def test_always_fail_decode():
    class _bytes(bytes):
        def decode(self, encoding="utf-8", errors="strict"):
            raise UnicodeDecodeError(
                encoding, self, 0, 0, "Failing intentionally for testing"
            )

    with pytest.raises(CharacterSetDecodeError):
        cmd._try_decode(_bytes())



from __future__ import annotations

from typing import TYPE_CHECKING

import yaml

from commitizen.exceptions import InvalidConfigurationError
from commitizen.git import smart_open

from .base_config import BaseConfig

if TYPE_CHECKING:
    import sys
    from pathlib import Path

    # Self is Python 3.11+ but backported in typing-extensions
    if sys.version_info < (3, 11):
        from typing_extensions import Self
    else:
        from typing import Self


class YAMLConfig(BaseConfig):
    def __init__(self, *, data: bytes | str, path: Path) -> None:
        super().__init__()
        self.path = path
        self._parse_setting(data)

    def init_empty_config_content(self) -> None:
        with smart_open(
            self.path, "a", encoding=self._settings["encoding"]
        ) as json_file:
            yaml.dump({"commitizen": {}}, json_file, explicit_start=True)

    def _parse_setting(self, data: bytes | str) -> None:
        """We expect to have a section in cz.yaml looking like

        ```
        commitizen:
          name: cz_conventional_commits
        ```
        """
        import yaml.scanner

        try:
            doc = yaml.safe_load(data)
        except yaml.YAMLError as e:
            raise InvalidConfigurationError(f"Failed to parse {self.path}: {e}")

        try:
            self.settings.update(doc["commitizen"])
        except (KeyError, TypeError):
            self.is_empty_config = True

    def set_key(self, key: str, value: object) -> Self:
        with open(self.path, "rb") as yaml_file:
            config_doc = yaml.load(yaml_file, Loader=yaml.FullLoader)

        config_doc["commitizen"][key] = value
        with smart_open(
            self.path, "w", encoding=self._settings["encoding"]
        ) as yaml_file:
            yaml.dump(config_doc, yaml_file, explicit_start=True)

        return self



from __future__ import annotations

from textwrap import dedent
from typing import TYPE_CHECKING

import pytest

from commitizen.changelog import Metadata
from commitizen.changelog_formats.restructuredtext import (
    RestructuredText,
    _is_overlined_title,
    _is_underlined_title,
)

if TYPE_CHECKING:
    from pathlib import Path

    from _pytest.mark.structures import ParameterSet

    from commitizen.config.base_config import BaseConfig


CASES: list[ParameterSet] = []


def case(
    id: str,
    content: str,
    latest_version: str | None = None,
    latest_version_position: int | None = None,
    latest_version_tag: str | None = None,
    unreleased_start: int | None = None,
    unreleased_end: int | None = None,
):
    CASES.append(
        pytest.param(
            dedent(content).strip(),
            Metadata(
                latest_version=latest_version,
                latest_version_tag=latest_version_tag,
                latest_version_position=latest_version_position,
                unreleased_start=unreleased_start,
                unreleased_end=unreleased_end,
            ),
            id=id,
        )
    )


case(
    "underlined title with intro and unreleased section",
    """
    Changelog
    #########

    All notable changes to this project will be documented in this file.

    The format is based on `Keep a Changelog <https://keepachangelog.com/en/1.0.0/>`,
    and this project adheres to `Semantic Versioning <https://semver.org/spec/v2.0.0.html>`.

    Unreleased
    ==========
    * Start using "changelog" over "change log" since it's the common usage.

    1.0.0 - 2017-06-20
    ==================
    Added
    -----
    * New visual identity by `@tylerfortune8 <https://github.com/tylerfortune8>`.
    * Version navigation.
    """,
    latest_version="1.0.0",
    latest_version_position=12,
    unreleased_start=8,
    unreleased_end=12,
)

case(
    "unreleased section without preamble",
    """
    Unreleased
    ==========
    * Start using "changelog" over "change log" since it's the common usage.

    1.2.0
    =====
    """,
    latest_version="1.2.0",
    latest_version_position=4,
    unreleased_start=0,
    unreleased_end=4,
)

case(
    "basic underlined titles with v-prefixed version",
    """
    Unreleased
    ==========

    v1.0.0
    ======
    """,
    latest_version="1.0.0",
    latest_version_tag="v1.0.0",
    latest_version_position=3,
    unreleased_start=0,
    unreleased_end=3,
)

case(
    "intermediate section in unreleased",
    """
    Unreleased
    ==========

    intermediate
    ------------

    1.0.0
    =====
    """,
    latest_version="1.0.0",
    latest_version_position=6,
    unreleased_start=0,
    unreleased_end=6,
)

case(
    "weird section with different level than versions",
    """
    Unreleased
    ##########

    1.0.0
    =====
    """,
    latest_version="1.0.0",
    latest_version_position=3,
    unreleased_start=0,
    unreleased_end=3,
)

case(
    "overlined title without release and intro",
    """
    ==========
    Unreleased
    ==========
    * Start using "changelog" over "change log" since it's the common usage.
    """,
    unreleased_start=0,
    unreleased_end=4,
)

case(
    "underlined title with date",
    """
    1.0.0 - 2017-06-20
    ==================
    """,
    latest_version="1.0.0",
    latest_version_position=0,
)


UNDERLINED_TITLES = (
    """
    title
    =====
    """,
    """
    title
    ======
    """,
    """
    title
    #####
    """,
    """
    title
    .....
    """,
    """
    title
    !!!!!
    """,
)

NOT_UNDERLINED_TITLES = (
    """
    title
    =.=.=
    """,
    """
    title
    ====
    """,
    """
    title
    aaaaa
    """,
    """
    title

    """,
)


OVERLINED_TITLES = (
    """
    =====
    title
    =====
    """,
    """
    ======
    title
    ======
    """,
    """
    #####
    title
    #####
    """,
    """
    .....
    title
    .....
    """,
)

NOT_OVERLINED_TITLES = (
    """
    ====
    title
    =====
    """,
    """
    =====
    title
    ====
    """,
    """
    ====
    title
    ====
    """,
    """
    =====
    title
    #####
    """,
    """
    #####
    title
    =====
    """,
    """
    =.=.=
    title
    =====
    """,
    """
    =====
    title
    =.=.=
    """,
    """

    title
    =====
    """,
    """
    =====
    title

    """,
    """
    aaaaa
    title
    aaaaa
    """,
)

CHANGELOG = """
Changelog
    #########

    All notable changes to this project will be documented in this file.

    The format is based on `Keep a Changelog <https://keepachangelog.com/en/1.0.0/>`,
    and this project adheres to `Semantic Versioning <https://semver.org/spec/v2.0.0.html>`.

    Unreleased
    ==========
    * Start using "changelog" over "change log" since it's the common usage.

    {tag_formatted_version} - 2017-06-20
    {underline}
    Added
    -----
    * New visual identity by `@tylerfortune8 <https://github.com/tylerfortune8>`.
    * Version navigation.
""".strip()


@pytest.fixture
def format(config: BaseConfig) -> RestructuredText:
    return RestructuredText(config)


@pytest.fixture
def format_with_tags(config: BaseConfig, request) -> RestructuredText:
    config.settings["tag_format"] = request.param
    config.settings["legacy_tag_formats"] = ["legacy-${version}"]
    return RestructuredText(config)


@pytest.mark.parametrize("content, expected", CASES)
def test_get_metadata(
    tmp_path: Path, format: RestructuredText, content: str, expected: Metadata
):
    changelog = tmp_path / format.default_changelog_file
    changelog.write_text(content)

    assert format.get_metadata(str(changelog)) == expected


@pytest.mark.parametrize(
    "text, expected",
    [(text, True) for text in UNDERLINED_TITLES]
    + [(text, False) for text in NOT_UNDERLINED_TITLES],
)
def test_is_underlined_title(text: str, expected: bool):
    _, first, second = dedent(text).splitlines()
    assert _is_underlined_title(first, second) is expected


@pytest.mark.parametrize(
    "text, expected",
    [(text, True) for text in OVERLINED_TITLES]
    + [(text, False) for text in NOT_OVERLINED_TITLES],
)
def test_is_overlined_title(text: str, expected: bool):
    _, first, second, third = dedent(text).splitlines()

    assert _is_overlined_title(first, second, third) is expected


@pytest.mark.parametrize(
    "format_with_tags, tag_string, expected, ",
    (
        pytest.param("${version}-example", "1.0.0-example", "1.0.0"),
        pytest.param("${version}", "1.0.0", "1.0.0"),
        pytest.param("${version}example", "1.0.0example", "1.0.0"),
        pytest.param("example${version}", "example1.0.0", "1.0.0"),
        pytest.param("example-${version}", "example-1.0.0", "1.0.0"),
        pytest.param("example-${major}-${minor}-${patch}", "example-1-0-0", "1.0.0"),
        pytest.param("example-${major}-${minor}", "example-1-0-0", None),
        pytest.param(
            "${major}-${minor}-${patch}-${prerelease}-example",
            "1-0-0-rc1-example",
            "1.0.0-rc1",
        ),
        pytest.param(
            "${major}-${minor}-${patch}-${prerelease}${devrelease}-example",
            "1-0-0-a1.dev1-example",
            "1.0.0-a1.dev1",
        ),
        pytest.param("new-${version}", "legacy-1.0.0", "1.0.0"),
    ),
    indirect=["format_with_tags"],
)
def test_get_metadata_custom_tag_format(
    tmp_path: Path,
    format_with_tags: RestructuredText,
    tag_string: str,
    expected: Metadata,
):
    content = CHANGELOG.format(
        tag_formatted_version=tag_string,
        underline="=" * len(tag_string) + "=============",
    )
    changelog = tmp_path / format_with_tags.default_changelog_file
    changelog.write_text(content)

    assert format_with_tags.get_metadata(str(changelog)).latest_version == expected



from __future__ import annotations

import contextlib
import os
import shutil
import subprocess
import tempfile
from typing import TYPE_CHECKING, TypedDict

import questionary

from commitizen import factory, git, out
from commitizen.cz.exceptions import CzException
from commitizen.cz.utils import get_backup_file_path
from commitizen.exceptions import (
    CommitError,
    CommitMessageLengthExceededError,
    CustomError,
    DryRunExit,
    NoAnswersError,
    NoCommitBackupError,
    NotAGitProjectError,
    NotAllowed,
    NothingToCommitError,
)
from commitizen.git import smart_open

if TYPE_CHECKING:
    from pathlib import Path

    from commitizen.config import BaseConfig


class CommitArgs(TypedDict, total=False):
    all: bool
    dry_run: bool
    edit: bool
    extra_cli_args: str
    message_length_limit: int | None
    no_retry: bool
    signoff: bool
    write_message_to_file: Path | None
    retry: bool


class Commit:
    """Show prompt for the user to create a guided commit."""

    def __init__(self, config: BaseConfig, arguments: CommitArgs) -> None:
        if not git.is_git_project():
            raise NotAGitProjectError()

        self.config: BaseConfig = config
        self.cz = factory.committer_factory(self.config)
        self.arguments = arguments
        self.backup_file_path = get_backup_file_path()

    def _read_backup_message(self) -> str | None:
        # Check the commit backup file exists
        if not self.backup_file_path.is_file():
            return None

        # Read commit message from backup
        with open(
            self.backup_file_path, encoding=self.config.settings["encoding"]
        ) as f:
            return f.read().strip()

    def _get_message_by_prompt_commit_questions(self) -> str:
        # Prompt user for the commit message
        questions = self.cz.questions()
        for question in (q for q in questions if q["type"] == "list"):
            question["use_shortcuts"] = self.config.settings["use_shortcuts"]
        try:
            answers = questionary.prompt(questions, style=self.cz.style)
        except ValueError as err:
            root_err = err.__context__
            if isinstance(root_err, CzException):
                raise CustomError(str(root_err))
            raise err

        if not answers:
            raise NoAnswersError()

        message = self.cz.message(answers)
        if limit := self.arguments.get(
            "message_length_limit", self.config.settings.get("message_length_limit", 0)
        ):
            self._validate_subject_length(message=message, length_limit=limit)

        return message

    def _validate_subject_length(self, *, message: str, length_limit: int) -> None:
        # By the contract, message_length_limit is set to 0 for no limit
        subject = message.partition("\n")[0].strip()
        if len(subject) > length_limit:
            raise CommitMessageLengthExceededError(
                f"Length of commit message exceeds limit ({len(subject)}/{length_limit}), subject: '{subject}'"
            )

    def manual_edit(self, message: str) -> str:
        editor = git.get_core_editor()
        if editor is None:
            raise RuntimeError("No 'editor' value given and no default available.")
        exec_path = shutil.which(editor)
        if exec_path is None:
            raise RuntimeError(f"Editor '{editor}' not found.")
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as file:
            file.write(message)
        file_path = file.name
        argv = [exec_path, file_path]
        subprocess.call(argv)
        with open(file_path) as temp_file:
            message = temp_file.read().strip()
        os.unlink(file.name)
        return message

    def _get_message(self) -> str:
        if self.arguments.get("retry"):
            commit_message = self._read_backup_message()
            if commit_message is None:
                raise NoCommitBackupError()
            return commit_message

        if (
            self.config.settings.get("retry_after_failure")
            and not self.arguments.get("no_retry")
            and (backup_message := self._read_backup_message())
        ):
            return backup_message
        return self._get_message_by_prompt_commit_questions()

    def __call__(self) -> None:
        extra_args = self.arguments.get("extra_cli_args", "")
        dry_run = bool(self.arguments.get("dry_run"))
        write_message_to_file = self.arguments.get("write_message_to_file")
        signoff = bool(self.arguments.get("signoff"))

        if signoff:
            out.warn(
                "Deprecated warning: `cz commit -s` is deprecated and will be removed in v5, please use `cz commit -- -s` instead."
            )

        if self.arguments.get("all"):
            git.add("-u")

        if git.is_staging_clean() and not (dry_run or "--allow-empty" in extra_args):
            raise NothingToCommitError("No files added to staging!")

        if write_message_to_file is not None and write_message_to_file.is_dir():
            raise NotAllowed(f"{write_message_to_file} is a directory")

        commit_message = self._get_message()
        if self.arguments.get("edit"):
            commit_message = self.manual_edit(commit_message)

        out.info(f"\n{commit_message}\n")

        if write_message_to_file:
            with smart_open(
                write_message_to_file, "w", encoding=self.config.settings["encoding"]
            ) as file:
                file.write(commit_message)

        if dry_run:
            raise DryRunExit()

        if self.config.settings["always_signoff"] or signoff:
            extra_args = f"{extra_args} -s".strip()

        c = git.commit(commit_message, args=extra_args)
        if c.return_code != 0:
            out.error(c.err)

            # Create commit backup
            with smart_open(
                self.backup_file_path, "w", encoding=self.config.settings["encoding"]
            ) as f:
                f.write(commit_message)

            raise CommitError()

        if any(s in c.out for s in ("nothing added", "no changes added to commit")):
            out.error(c.out)
            return

        with contextlib.suppress(FileNotFoundError):
            self.backup_file_path.unlink()
        out.write(c.err)
        out.write(c.out)
        out.success("Commit successful!")



from __future__ import annotations

import re
import sys
from typing import TYPE_CHECKING, TypedDict

from commitizen import factory, git, out
from commitizen.exceptions import (
    InvalidCommandArgumentError,
    InvalidCommitMessageError,
    NoCommitsFoundError,
)

if TYPE_CHECKING:
    from commitizen.config import BaseConfig


class CheckArgs(TypedDict, total=False):
    commit_msg_file: str
    commit_msg: str
    rev_range: str
    allow_abort: bool
    message_length_limit: int | None
    allowed_prefixes: list[str]
    message: str
    use_default_range: bool


class Check:
    """Check if the current commit msg matches the commitizen format."""

    def __init__(self, config: BaseConfig, arguments: CheckArgs, *args: object) -> None:
        """Initial check command.

        Args:
            config: The config object required for the command to perform its action
            arguments: All the flags provided by the user
            cwd: Current work directory
        """
        self.commit_msg_file = arguments.get("commit_msg_file")
        self.commit_msg = arguments.get("message")
        self.rev_range = arguments.get("rev_range")
        self.allow_abort = bool(
            arguments.get("allow_abort", config.settings["allow_abort"])
        )

        self.use_default_range = bool(arguments.get("use_default_range"))
        self.max_msg_length = arguments.get(
            "message_length_limit", config.settings.get("message_length_limit", None)
        )

        # we need to distinguish between None and [], which is a valid value
        allowed_prefixes = arguments.get("allowed_prefixes")
        self.allowed_prefixes: list[str] = (
            allowed_prefixes
            if allowed_prefixes is not None
            else config.settings["allowed_prefixes"]
        )

        num_exclusive_args_provided = sum(
            arg is not None
            for arg in (
                self.commit_msg_file,
                self.commit_msg,
                self.rev_range,
            )
        )

        if num_exclusive_args_provided > 1:
            raise InvalidCommandArgumentError(
                "Only one of --rev-range, --message, and --commit-msg-file is permitted by check command! "
                "See 'cz check -h' for more information"
            )

        if num_exclusive_args_provided == 0 and not sys.stdin.isatty():
            self.commit_msg = sys.stdin.read()

        self.config: BaseConfig = config
        self.cz = factory.committer_factory(self.config)

    def __call__(self) -> None:
        """Validate if commit messages follows the conventional pattern.

        Raises:
            InvalidCommitMessageError: if the commit provided does not follow the conventional pattern
            NoCommitsFoundError: if no commit is found with the given range
        """
        commits = self._get_commits()
        if not commits:
            raise NoCommitsFoundError(f"No commit found with range: '{self.rev_range}'")

        pattern = re.compile(self.cz.schema_pattern())
        invalid_commits = [
            (commit, check.errors)
            for commit in commits
            if not (
                check := self.cz.validate_commit_message(
                    commit_msg=commit.message,
                    pattern=pattern,
                    allow_abort=self.allow_abort,
                    allowed_prefixes=self.allowed_prefixes,
                    max_msg_length=self.max_msg_length,
                    commit_hash=commit.rev,
                )
            ).is_valid
        ]

        if invalid_commits:
            raise InvalidCommitMessageError(
                self.cz.format_exception_message(invalid_commits)
            )
        out.success("Commit validation: successful!")

    def _get_commit_message(self) -> str | None:
        if self.commit_msg_file is None:
            # Get commit message from command line (--message)
            return self.commit_msg

        with open(
            self.commit_msg_file, encoding=self.config.settings["encoding"]
        ) as commit_file:
            # Get commit message from file (--commit-msg-file)
            return commit_file.read()

    def _get_commits(self) -> list[git.GitCommit]:
        if (msg := self._get_commit_message()) is not None:
            return [git.GitCommit(rev="", title="", body=self._filter_comments(msg))]

        # Get commit messages from git log (--rev-range)
        return git.get_commits(
            git.get_default_branch() if self.use_default_range else None,
            self.rev_range,
        )

    @staticmethod
    def _filter_comments(msg: str) -> str:
        """Filter the commit message by removing comments.

        When using `git commit --verbose`, we exclude the diff that is going to
        generated, like the following example:

        ```bash
        ...
        # ------------------------ >8 ------------------------
        # Do not modify or remove the line above.
        # Everything below it will be ignored.
        diff --git a/... b/...
        ...
        ```

        Args:
            msg: The commit message to filter.

        Returns:
            The filtered commit message without comments.
        """

        lines: list[str] = []
        for line in msg.split("\n"):
            if "# ------------------------ >8 ------------------------" in line:
                break
            if not line.startswith("#"):
                lines.append(line)
        return "\n".join(lines)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from commitizen import defaults, git
from commitizen.config.factory import create_config
from commitizen.exceptions import ConfigFileIsEmpty, ConfigFileNotFound

from .base_config import BaseConfig

if TYPE_CHECKING:
    from collections.abc import Generator


def _resolve_config_paths(filepath: str | None = None) -> Generator[Path, None, None]:
    if filepath is not None:
        out_path = Path(filepath)
        if not out_path.exists():
            raise ConfigFileNotFound()

        yield out_path
        return

    git_project_root = git.find_git_project_root()
    cfg_search_paths = [Path(".")]
    if git_project_root:
        cfg_search_paths.append(git_project_root)

    for path in cfg_search_paths:
        for filename in defaults.CONFIG_FILES:
            out_path = path / Path(filename)
            if out_path.exists():
                yield out_path


def read_cfg(filepath: str | None = None) -> BaseConfig:
    for filename in _resolve_config_paths(filepath):
        with open(filename, "rb") as f:
            data: bytes = f.read()

        conf = create_config(data=data, path=filename)
        if not conf.is_empty_config:
            return conf

        if filepath is not None:
            raise ConfigFileIsEmpty()

    return BaseConfig()



from __future__ import annotations

import argparse
import logging
import sys
from copy import deepcopy
from functools import partial
from pathlib import Path
from types import TracebackType
from typing import TYPE_CHECKING, cast

import argcomplete
from decli import cli

from commitizen import commands, config, out, version_schemes
from commitizen.exceptions import (
    CommitizenException,
    ExitCode,
    ExpectedExit,
    InvalidCommandArgumentError,
    NoCommandFoundError,
)

logger = logging.getLogger(__name__)


class ParseKwargs(argparse.Action):
    """
    Parse arguments in the for `key=value`.

    Quoted strings are automatically unquoted.
    Can be submitted multiple times:

    ex:
        -k key=value -k double-quotes="value" -k single-quotes='value'

        will result in

        namespace["opt"] == {
            "key": "value",
            "double-quotes": "value",
            "single-quotes": "value",
        }
    """

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: object,
        option_string: str | None = None,
    ) -> None:
        if not isinstance(values, str):
            return

        key, sep, value = values.partition("=")
        if not key or not sep:
            raise InvalidCommandArgumentError(
                f"Option {option_string} expect a key=value format"
            )
        kwargs = getattr(namespace, self.dest, None) or {}
        kwargs[key] = value.strip("'\"")
        setattr(namespace, self.dest, kwargs)


tpl_arguments = (
    {
        "name": ["--template", "-t"],
        "help": (
            "changelog template file name (relative to the current working directory)"
        ),
    },
    {
        "name": ["--extra", "-e"],
        "action": ParseKwargs,
        "dest": "extras",
        "metavar": "EXTRA",
        "help": "a changelog extra variable (in the form 'key=value')",
    },
)

data = {
    "prog": "cz",
    "description": (
        "Commitizen is a powerful release management tool that helps teams maintain consistent and meaningful commit messages while automating version management.\n"
        "For more information, please visit https://commitizen-tools.github.io/commitizen"
    ),
    "formatter_class": argparse.RawDescriptionHelpFormatter,
    "arguments": [
        {
            "name": "--config",
            "help": "the path of configuration file",
        },
        {"name": "--debug", "action": "store_true", "help": "use debug mode"},
        {
            "name": ["-n", "--name"],
            "help": "use the given commitizen (default: cz_conventional_commits)",
        },
        {
            "name": ["-nr", "--no-raise"],
            "type": str,
            "required": False,
            "help": "comma separated error codes that won't raise error, e.g: cz -nr 1,2,3 bump. See codes at https://commitizen-tools.github.io/commitizen/exit_codes/",
        },
    ],
    "subcommands": {
        "title": "commands",
        "required": True,
        "commands": [
            {
                "name": ["init"],
                "description": "init commitizen configuration",
                "help": "init commitizen configuration",
                "func": commands.Init,
            },
            {
                "name": ["commit", "c"],
                "description": "create new commit",
                "help": "create new commit",
                "func": commands.Commit,
                "arguments": [
                    {
                        "name": ["--retry"],
                        "action": "store_true",
                        "help": "retry last commit",
                    },
                    {
                        "name": ["--no-retry"],
                        "action": "store_true",
                        "default": False,
                        "help": "skip retry if retry_after_failure is set to true",
                    },
                    {
                        "name": "--dry-run",
                        "action": "store_true",
                        "help": "show output to stdout, no commit, no modified files",
                    },
                    {
                        "name": "--write-message-to-file",
                        "type": Path,
                        "metavar": "FILE_PATH",
                        "help": "write message to file before committing (can be combined with --dry-run)",
                    },
                    {
                        "name": ["-s", "--signoff"],
                        "action": "store_true",
                        "help": "Deprecated, use 'cz commit -- -s' instead",
                    },
                    {
                        "name": ["-a", "--all"],
                        "action": "store_true",
                        "help": "Tell the command to automatically stage files that have been modified and deleted, but new files you have not told Git about are not affected.",
                    },
                    {
                        "name": ["-e", "--edit"],
                        "action": "store_true",
                        "default": False,
                        "help": "edit the commit message before committing",
                    },
                    {
                        "name": ["-l", "--message-length-limit"],
                        "type": int,
                        "help": "length limit of the commit message; 0 for no limit",
                    },
                    {
                        "name": ["--"],
                        "action": "store_true",
                        "dest": "double_dash",
                        "help": "Positional arguments separator (recommended)",
                    },
                ],
            },
            {
                "name": "ls",
                "description": "show available commitizens",
                "help": "show available commitizens",
                "func": commands.ListCz,
            },
            {
                "name": "example",
                "description": "show commit example",
                "help": "show commit example",
                "func": commands.Example,
            },
            {
                "name": "info",
                "description": "show information about the cz",
                "help": "show information about the cz",
                "func": commands.Info,
            },
            {
                "name": "schema",
                "description": "show commit schema",
                "help": "show commit schema",
                "func": commands.Schema,
            },
            {
                "name": "bump",
                "description": "bump semantic version based on the git log",
                "help": "bump semantic version based on the git log",
                "func": commands.Bump,
                "arguments": [
                    {
                        "name": "--dry-run",
                        "action": "store_true",
                        "help": "show output to stdout, no commit, no modified files",
                    },
                    {
                        "name": "--files-only",
                        "action": "store_true",
                        "help": "bump version in the files from the config",
                    },
                    {
                        "name": "--local-version",
                        "action": "store_true",
                        "help": "bump only the local version portion",
                    },
                    {
                        "name": ["--changelog", "-ch"],
                        "action": "store_true",
                        "default": False,
                        "help": "generate the changelog for the newest version",
                    },
                    {
                        "name": ["--no-verify"],
                        "action": "store_true",
                        "default": False,
                        "help": "this option bypasses the pre-commit and commit-msg hooks",
                    },
                    {
                        "name": "--yes",
                        "action": "store_true",
                        "help": "accept automatically questions done",
                    },
                    {
                        "name": "--tag-format",
                        "help": (
                            "the format used to tag the commit and read it, "
                            "use it in existing projects, "
                            "wrap around simple quotes"
                        ),
                    },
                    {
                        "name": "--bump-message",
                        "help": (
                            "template used to create the release commit, "
                            "useful when working with CI"
                        ),
                    },
                    {
                        "name": ["--prerelease", "-pr"],
                        "help": "choose type of prerelease",
                        "choices": ["alpha", "beta", "rc"],
                    },
                    {
                        "name": ["--devrelease", "-d"],
                        "help": "specify non-negative integer for dev. release",
                        "type": int,
                    },
                    {
                        "name": ["--increment"],
                        "help": "manually specify the desired increment",
                        "choices": ["MAJOR", "MINOR", "PATCH"],
                        "type": str.upper,
                    },
                    {
                        "name": ["--increment-mode"],
                        "choices": ["linear", "exact"],
                        "default": "linear",
                        "help": (
                            "set the method by which the new version is chosen. "
                            "'linear' (default) guesses the next version based on typical linear version progression, "
                            "such that bumping of a pre-release with lower precedence than the current pre-release "
                            "phase maintains the current phase of higher precedence. "
                            "'exact' applies the changes that have been specified (or determined from the commit log) "
                            "without interpretation, such that the increment and pre-release are always honored"
                        ),
                    },
                    {
                        "name": ["--check-consistency", "-cc"],
                        "help": (
                            "check consistency among versions defined in "
                            "commitizen configuration and version_files"
                        ),
                        "action": "store_true",
                    },
                    {
                        "name": ["--annotated-tag", "-at"],
                        "help": "create annotated tag instead of lightweight one",
                        "action": "store_true",
                    },
                    {
                        "name": ["--annotated-tag-message", "-atm"],
                        "help": "create annotated tag message",
                        "type": str,
                    },
                    {
                        "name": ["--gpg-sign", "-s"],
                        "help": "sign tag instead of lightweight one",
                        "default": False,
                        "action": "store_true",
                    },
                    {
                        "name": ["--changelog-to-stdout"],
                        "action": "store_true",
                        "default": False,
                        "help": "Output changelog to the stdout",
                    },
                    {
                        "name": ["--git-output-to-stderr"],
                        "action": "store_true",
                        "default": False,
                        "help": "Redirect git output to stderr",
                    },
                    {
                        "name": ["--retry"],
                        "action": "store_true",
                        "default": False,
                        "help": "retry commit if it fails the 1st time",
                    },
                    {
                        "name": ["--major-version-zero"],
                        "action": "store_true",
                        "default": None,
                        "help": "keep major version at zero, even for breaking changes",
                    },
                    *deepcopy(tpl_arguments),
                    {
                        "name": "--file-name",
                        "help": "file name of changelog (default: 'CHANGELOG.md')",
                    },
                    {
                        "name": ["--prerelease-offset"],
                        "type": int,
                        "default": None,
                        "help": "start pre-releases with this offset",
                    },
                    {
                        "name": ["--version-scheme"],
                        "help": "choose version scheme",
                        "default": None,
                        "choices": version_schemes.KNOWN_SCHEMES,
                    },
                    {
                        "name": ["--version-type"],
                        "help": "Deprecated, use --version-scheme instead",
                        "default": None,
                        "choices": version_schemes.KNOWN_SCHEMES,
                    },
                    {
                        "name": "manual_version",
                        "type": str,
                        "nargs": "?",
                        "help": "bump to the given version (e.g: 1.5.3)",
                        "metavar": "MANUAL_VERSION",
                    },
                    {
                        "name": ["--build-metadata"],
                        "help": "Add additional build-metadata to the version-number",
                        "default": None,
                    },
                    {
                        "name": ["--get-next"],
                        "action": "store_true",
                        "help": "Determine the next version and write to stdout",
                        "default": False,
                    },
                    {
                        "name": ["--allow-no-commit"],
                        "default": False,
                        "help": "bump version without eligible commits",
                        "action": "store_true",
                    },
                ],
            },
            {
                "name": ["changelog", "ch"],
                "description": (
                    "generate changelog (note that it will overwrite existing file)"
                ),
                "help": (
                    "generate changelog (note that it will overwrite existing file)"
                ),
                "func": commands.Changelog,
                "arguments": [
                    {
                        "name": "--dry-run",
                        "action": "store_true",
                        "default": False,
                        "help": "show changelog to stdout",
                    },
                    {
                        "name": "--file-name",
                        "help": "file name of changelog (default: 'CHANGELOG.md')",
                    },
                    {
                        "name": "--unreleased-version",
                        "help": (
                            "set the value for the new version (use the tag value), "
                            "instead of using unreleased"
                        ),
                    },
                    {
                        "name": "--incremental",
                        "action": "store_true",
                        "default": False,
                        "help": (
                            "generates changelog from last created version, "
                            "useful if the changelog has been manually modified"
                        ),
                    },
                    {
                        "name": "rev_range",
                        "type": str,
                        "nargs": "?",
                        "help": "generates changelog for the given version (e.g: 1.5.3) or version range (e.g: 1.5.3..1.7.9)",
                    },
                    {
                        "name": "--start-rev",
                        "default": None,
                        "help": (
                            "start rev of the changelog. "
                            "If not set, it will generate changelog from the start"
                        ),
                    },
                    {
                        "name": "--merge-prerelease",
                        "action": "store_true",
                        "default": False,
                        "help": (
                            "collect all changes from prereleases into next non-prerelease. "
                            "If not set, it will include prereleases in the changelog"
                        ),
                    },
                    {
                        "name": ["--version-scheme"],
                        "help": "choose version scheme",
                        "default": None,
                        "choices": version_schemes.KNOWN_SCHEMES,
                    },
                    {
                        "name": "--export-template",
                        "default": None,
                        "help": "Export the changelog template into this file instead of rendering it",
                    },
                    *deepcopy(tpl_arguments),
                    {
                        "name": "--tag-format",
                        "help": "The format of the tag, wrap around simple quotes",
                    },
                ],
            },
            {
                "name": ["check"],
                "description": "validates that a commit message matches the commitizen schema",
                "help": "validates that a commit message matches the commitizen schema",
                "func": commands.Check,
                "arguments": [
                    {
                        "name": "--commit-msg-file",
                        "help": (
                            "ask for the name of the temporal file that contains "
                            "the commit message. "
                            "Using it in a git hook script: MSG_FILE=$1"
                        ),
                        "exclusive_group": "group1",
                    },
                    {
                        "name": "--rev-range",
                        "help": "a range of git rev to check. e.g, master..HEAD",
                        "exclusive_group": "group1",
                    },
                    {
                        "name": ["-d", "--use-default-range"],
                        "action": "store_true",
                        "default": False,
                        "help": "check from the default branch to HEAD. e.g, refs/remotes/origin/master..HEAD",
                        "exclusive_group": "group1",
                    },
                    {
                        "name": ["-m", "--message"],
                        "help": "commit message that needs to be checked",
                        "exclusive_group": "group1",
                    },
                    {
                        "name": ["--allow-abort"],
                        "action": "store_true",
                        "default": False,
                        "help": "allow empty commit messages, which typically abort a commit",
                    },
                    {
                        "name": ["--allowed-prefixes"],
                        "nargs": "*",
                        "help": "allowed commit message prefixes. "
                        "If the message starts by one of these prefixes, "
                        "the message won't be checked against the regex",
                    },
                    {
                        "name": ["-l", "--message-length-limit"],
                        "type": int,
                        "help": "length limit of the commit message; 0 for no limit",
                    },
                ],
            },
            {
                "name": ["version"],
                "description": (
                    "get the version of the installed commitizen or the current project"
                    " (default: installed commitizen)"
                ),
                "help": (
                    "get the version of the installed commitizen or the current project"
                    " (default: installed commitizen)"
                ),
                "func": commands.Version,
                "arguments": [
                    {
                        "name": ["-r", "--report"],
                        "help": "get system information for reporting bugs",
                        "action": "store_true",
                        "exclusive_group": "group1",
                    },
                    {
                        "name": ["-p", "--project"],
                        "help": "get the version of the current project",
                        "action": "store_true",
                        "exclusive_group": "group1",
                    },
                    {
                        "name": ["-c", "--commitizen"],
                        "help": "get the version of the installed commitizen",
                        "action": "store_true",
                        "exclusive_group": "group1",
                    },
                    {
                        "name": ["-v", "--verbose"],
                        "help": (
                            "get the version of both the installed commitizen "
                            "and the current project"
                        ),
                        "action": "store_true",
                        "exclusive_group": "group1",
                    },
                    {
                        "name": ["--major"],
                        "help": "get just the major version. Need to be used with --project or --verbose.",
                        "action": "store_true",
                        "exclusive_group": "group2",
                    },
                    {
                        "name": ["--minor"],
                        "help": "get just the minor version. Need to be used with --project or --verbose.",
                        "action": "store_true",
                        "exclusive_group": "group2",
                    },
                ],
            },
        ],
    },
}


def commitizen_excepthook(
    type: type[BaseException],
    value: BaseException,
    traceback: TracebackType | None,
    debug: bool = False,
    no_raise: list[int] | None = None,
) -> None:
    traceback = traceback if isinstance(traceback, TracebackType) else None
    if not isinstance(value, CommitizenException):
        sys.__excepthook__(type, value, traceback)
        return

    if value.message:
        value.output_method(value.message)
    if debug:
        sys.__excepthook__(type, value, traceback)
    exit_code = value.exit_code
    if no_raise is not None and exit_code in no_raise:
        sys.exit(ExitCode.EXPECTED_EXIT)
    sys.exit(exit_code)


def parse_no_raise(comma_separated_no_raise: str) -> list[int]:
    """Convert the given string to exit codes.

    Receives digits and strings and outputs the parsed integer which
    represents the exit code found in exceptions.
    """

    def exit_code_from_str_or_skip(s: str) -> ExitCode | None:
        try:
            return ExitCode.from_str(s)
        except (KeyError, ValueError):
            out.warn(f"WARN: no_raise value `{s}` is not a valid exit code. Skipping.")
            return None

    return [
        code.value
        for s in comma_separated_no_raise.split(",")
        if (code := exit_code_from_str_or_skip(s)) is not None
    ]


if TYPE_CHECKING:

    class Args(argparse.Namespace):
        config: str | None = None
        debug: bool = False
        name: str | None = None
        no_raise: str | None = None  # comma-separated string, later parsed as list[int]
        report: bool = False
        project: bool = False
        commitizen: bool = False
        verbose: bool = False
        func: type[
            commands.Init  # init
            | commands.Commit  # commit (c)
            | commands.ListCz  # ls
            | commands.Example  # example
            | commands.Info  # info
            | commands.Schema  # schema
            | commands.Bump  # bump
            | commands.Changelog  # changelog (ch)
            | commands.Check  # check
            | commands.Version  # version
        ]


def main() -> None:
    parser: argparse.ArgumentParser = cli(data)
    argcomplete.autocomplete(parser)
    # Show help if no arg provided
    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        raise ExpectedExit()

    # This is for the command required constraint in 2.0
    try:
        args, unknown_args = parser.parse_known_args()
    except SystemExit as e:
        if e.code == 2:
            raise NoCommandFoundError()
        raise e

    arguments = vars(args)
    if unknown_args:
        # Raise error for extra-args without -- separation
        if "--" not in unknown_args:
            raise InvalidCommandArgumentError(
                f"Invalid commitizen arguments were found: `{' '.join(unknown_args)}`. "
                "Please use -- separator for extra git args"
            )
        # Raise error for extra-args before --
        elif unknown_args[0] != "--":
            pos = unknown_args.index("--")
            raise InvalidCommandArgumentError(
                f"Invalid commitizen arguments were found before -- separator: `{' '.join(unknown_args[:pos])}`. "
            )
        # Log warning for -- without any extra args
        elif len(unknown_args) == 1:
            logger.warning(
                "\nWARN: Incomplete commit command: received -- separator without any following git arguments\n"
            )
        extra_args = " ".join(unknown_args[1:])
        arguments["extra_cli_args"] = extra_args

    conf = config.read_cfg(args.config)
    args = cast("Args", args)
    if args.name:
        conf.update({"name": args.name})
    elif not conf.path:
        conf.update({"name": "cz_conventional_commits"})

    sys.excepthook = commitizen_excepthook
    if args.debug:
        logging.getLogger("commitizen").setLevel(logging.DEBUG)
        sys.excepthook = partial(sys.excepthook, debug=True)
    if args.no_raise:
        sys.excepthook = partial(sys.excepthook, no_raise=parse_no_raise(args.no_raise))

    args.func(conf, arguments)()  # type: ignore[arg-type]


if __name__ == "__main__":
    main()



import sys
from pathlib import Path
from textwrap import dedent

import pytest
from pytest_mock import MockFixture

from commitizen import bump, cli, cmd, exceptions

conversion = [
    (
        ("1.2.3", "1.3.0", "bump: $current_version -> $new_version [skip ci]"),
        "bump: 1.2.3 -> 1.3.0 [skip ci]",
    ),
    (("1.2.3", "1.3.0", None), "bump: version 1.2.3 → 1.3.0"),
    (("1.2.3", "1.3.0", "release $new_version"), "release 1.3.0"),
]


@pytest.mark.parametrize("test_input,expected", conversion)
def test_create_tag(test_input, expected):
    current_version, new_version, message_template = test_input
    new_tag = bump.create_commit_message(current_version, new_version, message_template)
    assert new_tag == expected


@pytest.mark.parametrize(
    "retry",
    (
        pytest.param(
            True,
            marks=pytest.mark.skipif(
                sys.version_info >= (3, 13),
                reason="mirrors-prettier is not supported with Python 3.13 or higher",
            ),
        ),
        False,
    ),
)
@pytest.mark.usefixtures("tmp_commitizen_project")
def test_bump_pre_commit_changelog(mocker: MockFixture, freezer, retry):
    freezer.move_to("2022-04-01")
    testargs = ["cz", "bump", "--changelog", "--yes"]
    if retry:
        testargs.append("--retry")
    else:
        pytest.xfail("it will fail because pre-commit will reformat CHANGELOG.md")
    mocker.patch.object(sys, "argv", testargs)
    # Configure prettier as a pre-commit hook
    Path(".pre-commit-config.yaml").write_text(
        dedent(
            """\
            repos:
              - repo: https://github.com/pre-commit/mirrors-prettier
                rev: v3.0.3
                hooks:
                - id: prettier
                  stages: [commit]
            """
        )
    )
    # Prettier inherits editorconfig
    Path(".editorconfig").write_text(
        dedent(
            """\
            [*]
            indent_size = 4
            """
        )
    )
    cmd.run("git add -A")
    cmd.run('git commit -m "fix: _test"')
    cmd.run("pre-commit install")
    cli.main()
    # Pre-commit fixed last line adding extra indent and "\" char
    assert Path("CHANGELOG.md").read_text() == dedent(
        """\
        ## 0.1.1 (2022-04-01)

        ### Fix

        -   \\_test
        """
    )


@pytest.mark.parametrize("retry", (True, False))
@pytest.mark.usefixtures("tmp_commitizen_project")
def test_bump_pre_commit_changelog_fails_always(mocker: MockFixture, freezer, retry):
    freezer.move_to("2022-04-01")
    testargs = ["cz", "bump", "--changelog", "--yes"]
    if retry:
        testargs.append("--retry")
    mocker.patch.object(sys, "argv", testargs)
    Path(".pre-commit-config.yaml").write_text(
        dedent(
            """\
            repos:
              - repo: local
                hooks:
                - id: forbid-changelog
                  name: changelogs are forbidden
                  entry: changelogs are forbidden
                  language: fail
                  files: CHANGELOG.md
            """
        )
    )
    cmd.run("git add -A")
    cmd.run('git commit -m "feat: forbid changelogs"')
    cmd.run("pre-commit install")
    with pytest.raises(exceptions.BumpCommitFailedError):
        cli.main()


@pytest.mark.usefixtures("tmp_commitizen_project")
def test_bump_with_build_metadata(mocker: MockFixture, freezer):
    def _add_entry(test_str: str, args: list):
        Path(test_str).write_text("")
        cmd.run("git add -A")
        cmd.run(f'git commit -m "fix: test-{test_str}"')
        cz_args = ["cz", "bump", "--changelog", "--yes"] + args
        mocker.patch.object(sys, "argv", cz_args)
        cli.main()

    freezer.move_to("2024-01-01")

    _add_entry("a", ["--build-metadata", "a.b.c"])
    _add_entry("b", [])
    _add_entry("c", ["--build-metadata", "alongmetadatastring"])
    _add_entry("d", [])

    # Pre-commit fixed last line adding extra indent and "\" char
    assert Path("CHANGELOG.md").read_text() == dedent(
        """\
        ## 0.1.4 (2024-01-01)

        ### Fix

        - test-d

        ## 0.1.3+alongmetadatastring (2024-01-01)

        ### Fix

        - test-c

        ## 0.1.2 (2024-01-01)

        ### Fix

        - test-b

        ## 0.1.1+a.b.c (2024-01-01)

        ### Fix

        - test-a
        """
    )



import os
from unittest.mock import call

import pytest
from pytest_mock import MockFixture

from commitizen import cmd, hooks
from commitizen.exceptions import RunHookError


def test_run(mocker: MockFixture):
    bump_hooks = ["pre_bump_hook", "pre_bump_hook_1"]

    cmd_run_mock = mocker.Mock()
    cmd_run_mock.return_value.return_code = 0
    mocker.patch.object(cmd, "run", cmd_run_mock)

    hooks.run(bump_hooks)

    cmd_run_mock.assert_has_calls(
        [
            call("pre_bump_hook", env=dict(os.environ)),
            call("pre_bump_hook_1", env=dict(os.environ)),
        ]
    )


def test_run_error(mocker: MockFixture):
    bump_hooks = ["pre_bump_hook", "pre_bump_hook_1"]

    cmd_run_mock = mocker.Mock()
    cmd_run_mock.return_value.return_code = 1
    mocker.patch.object(cmd, "run", cmd_run_mock)

    with pytest.raises(RunHookError):
        hooks.run(bump_hooks)


def test_format_env():
    result = hooks._format_env("TEST_", {"foo": "bar", "bar": "baz"})
    assert "TEST_FOO" in result and result["TEST_FOO"] == "bar"
    assert "TEST_BAR" in result and result["TEST_BAR"] == "baz"



"""Design

## Metadata CHANGELOG.md

1. Identify irrelevant information (possible: changelog title, first paragraph)
2. Identify Unreleased area
3. Identify latest version (to be able to write on top of it)

## Parse git log

1. get commits between versions
2. filter commits with the current cz rules
3. parse commit information
4. yield tree nodes
5. format tree nodes
6. produce full tree
7. generate changelog

Extra:
- [x] Generate full or partial changelog
- [x] Include in tree from file all the extra comments added manually
- [x] Add unreleased value
- [x] hook after message is parsed (add extra information like hyperlinks)
- [x] hook after changelog is generated (api calls)
- [x] add support for change_type maps
"""

from __future__ import annotations

import re
from collections import OrderedDict, defaultdict
from dataclasses import dataclass
from datetime import date
from itertools import chain
from typing import TYPE_CHECKING, Any

from deprecated import deprecated
from jinja2 import (
    BaseLoader,
    ChoiceLoader,
    Environment,
    FileSystemLoader,
    Template,
)

from commitizen.exceptions import InvalidConfigurationError, NoCommitsFoundError
from commitizen.tags import TagRules

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable, Mapping, MutableMapping, Sequence

    from commitizen.cz.base import ChangelogReleaseHook, MessageBuilderHook
    from commitizen.git import GitCommit, GitTag


@dataclass
class Metadata:
    """
    Metadata extracted from the changelog produced by a plugin
    """

    unreleased_start: int | None = None
    unreleased_end: int | None = None
    latest_version: str | None = None
    latest_version_position: int | None = None
    latest_version_tag: str | None = None

    def __post_init__(self) -> None:
        if self.latest_version and not self.latest_version_tag:
            # Test syntactic sugar
            # latest version tag is optional if same as latest version
            self.latest_version_tag = self.latest_version


def get_commit_tag(commit: GitCommit, tags: list[GitTag]) -> GitTag | None:
    return next((tag for tag in tags if tag.rev == commit.rev), None)


def generate_tree_from_commits(
    commits: list[GitCommit],
    tags: list[GitTag],
    commit_parser: str,
    changelog_pattern: str,
    unreleased_version: str | None = None,
    change_type_map: dict[str, str] | None = None,
    changelog_message_builder_hook: MessageBuilderHook | None = None,
    changelog_release_hook: ChangelogReleaseHook | None = None,
    rules: TagRules | None = None,
) -> Generator[dict[str, Any], None, None]:
    pat = re.compile(changelog_pattern)
    map_pat = re.compile(commit_parser, re.MULTILINE)
    body_map_pat = re.compile(commit_parser, re.MULTILINE | re.DOTALL)
    rules = rules or TagRules()

    # Check if the latest commit is not tagged

    current_tag = get_commit_tag(commits[0], tags) if commits else None
    current_tag_name = unreleased_version or "Unreleased"
    current_tag_date = (
        date.today().isoformat() if unreleased_version is not None else ""
    )

    used_tags: set[GitTag] = set()
    if current_tag:
        used_tags.add(current_tag)
        if current_tag.name:
            current_tag_name = current_tag.name
            current_tag_date = current_tag.date

    commit_tag: GitTag | None = None
    changes: dict = defaultdict(list)
    for commit in commits:
        if (
            (commit_tag := get_commit_tag(commit, tags))
            and commit_tag not in used_tags
            and rules.include_in_changelog(commit_tag)
        ):
            used_tags.add(commit_tag)
            release = {
                "version": current_tag_name,
                "date": current_tag_date,
                "changes": changes,
            }
            if changelog_release_hook:
                release = changelog_release_hook(release, commit_tag)
            yield release
            current_tag_name = commit_tag.name
            current_tag_date = commit_tag.date
            changes = defaultdict(list)

        if not pat.match(commit.message):
            continue

        # Process subject and body from commit message
        for message in chain(
            [map_pat.match(commit.message)],
            (body_map_pat.match(block) for block in commit.body.split("\n\n")),
        ):
            if message:
                process_commit_message(
                    changelog_message_builder_hook,
                    message,
                    commit,
                    changes,
                    change_type_map,
                )

    release = {
        "version": current_tag_name,
        "date": current_tag_date,
        "changes": changes,
    }
    if changelog_release_hook:
        release = changelog_release_hook(release, commit_tag)
    yield release


def process_commit_message(
    hook: MessageBuilderHook | None,
    parsed: re.Match[str],
    commit: GitCommit,
    ref_changes: MutableMapping[str | None, list],
    change_type_map: Mapping[str, str] | None = None,
) -> None:
    message: dict[str, Any] = {
        "sha1": commit.rev,
        "parents": commit.parents,
        "author": commit.author,
        "author_email": commit.author_email,
        **parsed.groupdict(),
    }

    processed_msg = hook(message, commit) if hook else message
    if not processed_msg:
        return

    messages = [processed_msg] if isinstance(processed_msg, dict) else processed_msg
    for msg in messages:
        change_type = msg.pop("change_type", None)
        if change_type_map:
            change_type = change_type_map.get(change_type, change_type)
        ref_changes[change_type].append(msg)


def generate_ordered_changelog_tree(
    tree: Iterable[Mapping[str, Any]], change_type_order: list[str]
) -> Generator[dict[str, Any], None, None]:
    if len(set(change_type_order)) != len(change_type_order):
        raise InvalidConfigurationError(
            f"Change types contain duplicated types ({change_type_order})"
        )

    for entry in tree:
        yield {
            **entry,
            "changes": _calculate_sorted_changes(change_type_order, entry["changes"]),
        }


def _calculate_sorted_changes(
    change_type_order: list[str], changes: Mapping[str, Any]
) -> OrderedDict[str, Any]:
    remaining_change_types = set(changes.keys()) - set(change_type_order)
    sorted_change_types = change_type_order + sorted(remaining_change_types)
    return OrderedDict((ct, changes[ct]) for ct in sorted_change_types if ct in changes)


def get_changelog_template(loader: BaseLoader, template: str) -> Template:
    loader = ChoiceLoader(
        [
            FileSystemLoader("."),
            loader,
        ]
    )
    env = Environment(loader=loader, trim_blocks=True)
    return env.get_template(template)


def render_changelog(
    tree: Iterable,
    loader: BaseLoader,
    template: str,
    **kwargs: Any,
) -> str:
    jinja_template = get_changelog_template(loader, template)
    changelog: str = jinja_template.render(tree=tree, **kwargs)
    return changelog


def incremental_build(
    new_content: str, lines: list[str], metadata: Metadata
) -> list[str]:
    """Takes the original lines and updates with new_content.

    The metadata governs how to remove the old unreleased section and where to place the
    new content.

    Args:
        lines: The lines from the changelog
        new_content: This should be placed somewhere in the lines
        metadata: Information about the changelog

    Returns:
        Updated lines
    """
    unreleased_start = metadata.unreleased_start
    unreleased_end = metadata.unreleased_end
    latest_version_position = metadata.latest_version_position

    skip = False
    output_lines: list[str] = []
    for index, line in enumerate(lines):
        if index == unreleased_start:
            skip = True
        elif index == unreleased_end:
            skip = False
            if (
                latest_version_position is None
                or latest_version_position > unreleased_end
            ):
                continue

        if skip:
            continue

        if index == latest_version_position:
            output_lines.extend([new_content, "\n"])
        output_lines.append(line)

    if latest_version_position is not None:
        return output_lines

    if output_lines and output_lines[-1].strip():
        # Ensure at least one blank line between existing and new content.
        output_lines.append("\n")
    output_lines.append(new_content)
    return output_lines


def get_next_tag_name_after_version(tags: Iterable[GitTag], version: str) -> str | None:
    it = iter(tag.name for tag in tags)
    for name in it:
        if name == version:
            return next(it, None)

    raise NoCommitsFoundError(f"Could not find a valid revision range. {version=}")


@deprecated(
    reason="This function is unused and will be removed in v5",
    version="5.0.0",
    category=DeprecationWarning,
)
def get_smart_tag_range(
    tags: Sequence[GitTag], newest: str, oldest: str | None = None
) -> list[GitTag]:
    """Smart because it finds the N+1 tag.

    This is because we need to find until the next tag
    """
    accumulator = []
    keep = False
    if not oldest:
        oldest = newest
    for index, tag in enumerate(tags):
        if tag.name == newest:
            keep = True
        if keep:
            accumulator.append(tag)
        if tag.name == oldest:
            keep = False
            try:
                accumulator.append(tags[index + 1])
            except IndexError:
                pass
            break
    return accumulator


def get_oldest_and_newest_rev(
    tags: Iterable[GitTag],
    version: str,
    rules: TagRules,
) -> tuple[str | None, str]:
    """Find the tags for the given version.

    `version` may come in different formats:
    - `0.1.0..0.4.0`: as a range
    - `0.3.0`: as a single version
    """
    oldest_version, sep, newest_version = version.partition("..")
    if not sep:
        newest_version = version
        oldest_version = ""

    def get_tag_name(v: str) -> str:
        if tag := rules.find_tag_for(tags, v):
            return tag.name
        raise NoCommitsFoundError("Could not find a valid revision range.")

    newest_tag_name = get_tag_name(newest_version)
    oldest_tag_name = get_tag_name(oldest_version) if oldest_version else None

    oldest_rev = get_next_tag_name_after_version(
        tags, oldest_tag_name or newest_tag_name
    )

    # Return None for oldest_rev if:
    # 1. The oldest tag is the last tag in the list and matches the requested oldest tag
    # 2. The oldest and the newest tag are the same
    if oldest_rev == newest_tag_name:
        return None, newest_tag_name
    return oldest_rev, newest_tag_name



from __future__ import annotations

import warnings
from logging import getLogger
from typing import TYPE_CHECKING, cast

import questionary

from commitizen import bump, factory, git, hooks, out
from commitizen.changelog_formats import get_changelog_format
from commitizen.commands.changelog import Changelog
from commitizen.defaults import Settings
from commitizen.exceptions import (
    BumpCommitFailedError,
    BumpTagFailedError,
    DryRunExit,
    ExpectedExit,
    InvalidManualVersion,
    NoCommitsFoundError,
    NoneIncrementExit,
    NoPatternMapError,
    NotAGitProjectError,
    NotAllowed,
)
from commitizen.providers import get_provider
from commitizen.tags import TagRules
from commitizen.version_schemes import (
    Increment,
    InvalidVersion,
    Prerelease,
    VersionProtocol,
    get_version_scheme,
)

if TYPE_CHECKING:
    from commitizen.config import BaseConfig

logger = getLogger("commitizen")


class BumpArgs(Settings, total=False):
    allow_no_commit: bool | None
    annotated_tag_message: str | None
    build_metadata: str | None
    changelog_to_stdout: bool
    changelog: bool
    check_consistency: bool
    devrelease: int | None
    dry_run: bool
    file_name: str
    files_only: bool | None
    get_next: bool  # TODO: maybe rename to `next_version_to_stdout`
    git_output_to_stderr: bool
    increment_mode: str
    increment: Increment | None
    local_version: bool
    manual_version: str | None
    no_verify: bool
    prerelease: Prerelease | None
    retry: bool
    yes: bool


class Bump:
    """Show prompt for the user to create a guided commit."""

    def __init__(self, config: BaseConfig, arguments: BumpArgs) -> None:
        if not git.is_git_project():
            raise NotAGitProjectError()

        self.config: BaseConfig = config
        self.arguments = arguments
        self.bump_settings = cast(
            "BumpArgs",
            {
                **config.settings,
                **{
                    k: v
                    for k in (
                        "annotated_tag_message",
                        "annotated_tag",
                        "bump_message",
                        "file_name",
                        "gpg_sign",
                        "increment_mode",
                        "increment",
                        "major_version_zero",
                        "prerelease_offset",
                        "prerelease",
                        "tag_format",
                        "template",
                    )
                    if (v := arguments.get(k)) is not None
                },
            },
        )
        self.cz = factory.committer_factory(self.config)
        self.changelog_flag = arguments["changelog"]
        self.changelog_to_stdout = arguments["changelog_to_stdout"]
        self.git_output_to_stderr = arguments["git_output_to_stderr"]
        self.no_verify = arguments["no_verify"]
        self.check_consistency = arguments["check_consistency"]
        self.retry = arguments["retry"]
        self.pre_bump_hooks = self.config.settings["pre_bump_hooks"]
        self.post_bump_hooks = self.config.settings["post_bump_hooks"]
        deprecated_version_type = arguments.get("version_type")
        if deprecated_version_type:
            warnings.warn(
                DeprecationWarning(
                    "`--version-type` parameter is deprecated and will be removed in v5. "
                    "Please use `--version-scheme` instead"
                )
            )
        self.scheme = get_version_scheme(
            self.config.settings, arguments["version_scheme"] or deprecated_version_type
        )
        self.file_name = arguments["file_name"] or self.config.settings.get(
            "changelog_file"
        )
        self.changelog_format = get_changelog_format(self.config, self.file_name)

        self.template = (
            arguments["template"]
            or self.config.settings.get("template")
            or self.changelog_format.template
        )
        self.extras = arguments["extras"]

    def _is_initial_tag(
        self, current_tag: git.GitTag | None, is_yes: bool = False
    ) -> bool:
        """Check if reading the whole git tree up to HEAD is needed."""
        if current_tag:
            return False
        if is_yes:
            return True

        out.info("No tag matching configuration could be found.")
        out.info(
            "Possible causes:\n"
            "- version in configuration is not the current version\n"
            "- tag_format or legacy_tag_formats is missing, check them using 'git tag --list'\n"
        )
        return bool(questionary.confirm("Is this the first tag created?").ask())

    def _find_increment(self, commits: list[git.GitCommit]) -> Increment | None:
        # Update the bump map to ensure major version doesn't increment.
        # self.cz.bump_map = defaults.bump_map_major_version_zero
        bump_map = (
            self.cz.bump_map_major_version_zero
            if self.bump_settings["major_version_zero"]
            else self.cz.bump_map
        )
        bump_pattern = self.cz.bump_pattern

        if not bump_map or not bump_pattern:
            raise NoPatternMapError(
                f"'{self.config.settings['name']}' rule does not support bump"
            )
        return bump.find_increment(commits, regex=bump_pattern, increments_map=bump_map)

    def _validate_arguments(self, current_version: VersionProtocol) -> None:
        errors: list[str] = []
        if self.arguments["manual_version"]:
            for val, option in (
                (self.arguments["increment"], "--increment"),
                (self.arguments["prerelease"], "--prerelease"),
                (self.arguments["devrelease"] is not None, "--devrelease"),
                (self.arguments["local_version"], "--local-version"),
                (self.arguments["build_metadata"], "--build-metadata"),
                (self.arguments["major_version_zero"], "--major-version-zero"),
            ):
                if val:
                    errors.append(f"{option} cannot be combined with MANUAL_VERSION")

        if self.arguments["major_version_zero"] and current_version.release[0]:
            errors.append(
                f"--major-version-zero is meaningless for current version {current_version}"
            )
        if self.arguments["build_metadata"] and self.arguments["local_version"]:
            errors.append("--local-version cannot be combined with --build-metadata")

        if errors:
            raise NotAllowed("\n".join(errors))

    def _resolve_increment_and_new_version(
        self, current_version: VersionProtocol, current_tag: git.GitTag | None
    ) -> tuple[Increment | None, VersionProtocol]:
        increment = self.arguments["increment"]
        if manual_version := self.arguments["manual_version"]:
            try:
                return increment, self.scheme(manual_version)
            except InvalidVersion as exc:
                raise InvalidManualVersion(
                    "[INVALID_MANUAL_VERSION]\n"
                    f"Invalid manual version: '{manual_version}'"
                ) from exc

        if increment is None:
            commits = git.get_commits(current_tag.name if current_tag else None)

            # No commits, there is no need to create an empty tag.
            # Unless we previously had a prerelease.
            if (
                not commits
                and not current_version.is_prerelease
                and not self.arguments["allow_no_commit"]
            ):
                raise NoCommitsFoundError("[NO_COMMITS_FOUND]\nNo new commits found.")

            increment = self._find_increment(commits)

        # It may happen that there are commits, but they are not eligible
        # for an increment, this generates a problem when using prerelease (#281)
        if (
            self.arguments["prerelease"]
            and increment is None
            and not current_version.is_prerelease
        ):
            raise NoCommitsFoundError(
                "[NO_COMMITS_FOUND]\n"
                "No commits found to generate a pre-release.\n"
                "To avoid this error, manually specify the type of increment with `--increment`"
            )

        # we create an empty PATCH increment for empty tag
        if increment is None and self.arguments["allow_no_commit"]:
            increment = "PATCH"

        return increment, current_version.bump(
            increment,
            prerelease=self.arguments["prerelease"],
            prerelease_offset=self.bump_settings["prerelease_offset"],
            devrelease=self.arguments["devrelease"],
            is_local_version=self.arguments["local_version"],
            build_metadata=self.arguments["build_metadata"],
            exact_increment=self.arguments["increment_mode"] == "exact",
        )

    def __call__(self) -> None:
        """Steps executed to bump."""
        provider = get_provider(self.config)
        current_version = self.scheme(provider.get_version())
        self._validate_arguments(current_version)

        next_version_to_stdout = self.arguments["get_next"]
        if next_version_to_stdout:
            for value, option in (
                (self.changelog_flag, "--changelog"),
                (self.changelog_to_stdout, "--changelog-to-stdout"),
            ):
                if value:
                    warnings.warn(f"{option} has no effect when used with --get-next")

        # If user specified changelog_to_stdout, they probably want the
        # changelog to be generated as well, this is the most intuitive solution
        self.changelog_flag = any(
            (
                self.changelog_flag,
                self.changelog_to_stdout,
                self.config.settings.get("update_changelog_on_bump"),
            )
        )

        rules = TagRules.from_settings(cast("Settings", self.bump_settings))
        current_tag = rules.find_tag_for(git.get_tags(), current_version)
        current_tag_version = (
            current_tag.name if current_tag else rules.normalize_tag(current_version)
        )

        is_initial = self._is_initial_tag(current_tag, self.arguments["yes"])

        increment, new_version = self._resolve_increment_and_new_version(
            current_version, current_tag
        )

        new_tag_version = rules.normalize_tag(new_version)
        if next_version_to_stdout:
            if increment is None and new_tag_version == current_tag_version:
                raise NoneIncrementExit(
                    "[NO_COMMITS_TO_BUMP]\n"
                    "The commits found are not eligible to be bumped"
                )
            out.write(str(new_version))
            raise DryRunExit()

        message = bump.create_commit_message(
            current_version, new_version, self.bump_settings["bump_message"]
        )
        # Report found information
        information = f"{message}\ntag to create: {new_tag_version}\n"
        if increment:
            information += f"increment detected: {increment}\n"

        if self.changelog_to_stdout:
            # When the changelog goes to stdout, we want to send
            # the bump information to stderr, this way the
            # changelog output can be captured
            out.diagnostic(information)
        else:
            out.write(information)

        if increment is None and new_tag_version == current_tag_version:
            raise NoneIncrementExit(
                "[NO_COMMITS_TO_BUMP]\nThe commits found are not eligible to be bumped"
            )

        updated_files: list[str] = []
        dry_run = self.arguments["dry_run"]
        if self.changelog_flag:
            changelog_args = {
                "unreleased_version": new_tag_version,
                "template": self.template,
                "extras": self.extras,
                "incremental": True,
                "dry_run": dry_run,
            }
            if self.changelog_to_stdout:
                changelog_cmd = Changelog(
                    self.config,
                    {**changelog_args, "dry_run": True},  # type: ignore[typeddict-item]
                )
                try:
                    changelog_cmd()
                except DryRunExit:
                    pass

            changelog_cmd = Changelog(
                self.config,
                {**changelog_args, "file_name": self.file_name},  # type: ignore[typeddict-item]
            )
            changelog_cmd()
            updated_files.append(changelog_cmd.file_name)

        # Do not perform operations over files or git.
        if dry_run:
            raise DryRunExit()

        updated_files.extend(
            bump.update_version_in_files(
                str(current_version),
                str(new_version),
                self.bump_settings["version_files"],
                check_consistency=self.check_consistency,
                encoding=self.config.settings["encoding"],
            )
        )

        provider.set_version(str(new_version))

        if self.pre_bump_hooks:
            hooks.run(
                self.pre_bump_hooks,
                _env_prefix="CZ_PRE_",
                is_initial=is_initial,
                current_version=str(current_version),
                current_tag_version=current_tag_version,
                new_version=new_version.public,
                new_tag_version=new_tag_version,
                message=message,
                increment=increment,
                changelog_file_name=changelog_cmd.file_name
                if self.changelog_flag
                else None,
            )

        if self.arguments["files_only"]:
            raise ExpectedExit()

        # FIXME: check if any changes have been staged
        git.add(*updated_files)
        c = git.commit(message, args=self._get_commit_args())
        if self.retry and c.return_code != 0 and self.changelog_flag:
            # Maybe pre-commit reformatted some files? Retry once
            logger.debug("1st git.commit error: %s", c.err)
            logger.info("1st commit attempt failed; retrying once")
            git.add(*updated_files)
            c = git.commit(message, args=self._get_commit_args())
        if c.return_code != 0:
            err = c.err.strip() or c.out
            raise BumpCommitFailedError(f'2nd git.commit error: "{err}"')

        for msg in (c.out, c.err):
            if msg:
                out_func = out.diagnostic if self.git_output_to_stderr else out.write
                out_func(msg)

        c = git.tag(
            new_tag_version,
            signed=any(
                (
                    self.bump_settings.get("gpg_sign"),
                    self.config.settings.get("gpg_sign"),
                )
            ),
            annotated=any(
                (
                    self.bump_settings.get("annotated_tag"),
                    self.config.settings.get("annotated_tag"),
                    self.bump_settings.get("annotated_tag_message"),
                )
            ),
            msg=self.bump_settings.get("annotated_tag_message", None),
            # TODO: also get from self.config.settings?
        )
        if c.return_code != 0:
            raise BumpTagFailedError(c.err)

        if self.post_bump_hooks:
            hooks.run(
                self.post_bump_hooks,
                _env_prefix="CZ_POST_",
                was_initial=is_initial,
                previous_version=str(current_version),
                previous_tag_version=current_tag_version,
                current_version=new_version.public,
                current_tag_version=new_tag_version,
                message=message,
                increment=increment,
                changelog_file_name=changelog_cmd.file_name
                if self.changelog_flag
                else None,
            )

        # TODO: For v3 output this only as diagnostic and remove this if
        if self.changelog_to_stdout:
            out.diagnostic("Done!")
        else:
            out.success("Done!")

    def _get_commit_args(self) -> str:
        commit_args = ["-a"]
        if self.no_verify:
            commit_args.append("--no-verify")
        return " ".join(commit_args)
